unit ComplEdt;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Complex, StdCtrls, RealEdit;

type
  TComplexEdit = class(TFrame)
    x: TRealEdit;
    Label1: TLabel;
    y: TRealEdit;
    procedure FrameExit(Sender: TObject);
  private
    { Private declarations }
    FCValue: TComplexRec;
    FNorm, FSqrNorm, FArg: Extended;
    FNewValue, FNoNorm, FNewNorm, FNoArg: Boolean;
  protected
    procedure SetCValue(Value: TComplexRec);
    procedure SetRe(Value: Extended);
    procedure SetIm(Value: Extended);
    procedure SetNorm(Value: Extended);
    procedure SetArg(Value: Extended);
    function GetCValue: TComplexRec;
    function GetRe: Extended;
    function GetIm: Extended;
    function GetNorm: Extended;
    function GetSqrNorm: Extended;
    function GetArg: Extended;
    function GetValueStr: String;
  public
    { Public declarations }
    property NewValue: Boolean read FNewValue;
    property CValue: TComplexRec read GetCValue write SetCValue;
    property Re: Extended read GetRe write SetRe;
    property Im: Extended read GetIm write SetIm;
    property Norm: Extended read GetNorm write SetNorm;
    property SqrNorm: Extended read GetSqrNorm;
    property Arg: Extended read GetArg write SetArg;
    property ValueStr: string read GetValueStr;
    procedure AssignValues(Source: TComplexEdit);
    procedure SetValues(NewX, NewY: Extended);
    procedure Inc(const Ix, Iy: Extended); overload;
    procedure Inc(const I: TComplexRec); overload;
    procedure Dec(const D: TComplexRec);
    procedure IncRe(const I: Extended);
    procedure IncIm(const I: Extended);
    procedure Scale(const Scal: Extended; KeepNorm: Boolean = False);
  end;

implementation

{$R *.DFM}

procedure TComplexEdit.SetValues(NewX, NewY: Extended);
begin
  if not FNewValue then with FCValue do
     FNewValue := (x <> NewX) or (y <> NewY);
  if FNewValue then
     begin x.Value := NewX; y.Value := NewY end
end;

procedure TComplexEdit.SetCValue(Value: TComplexRec);
begin with Value do SetValues(x, y) end;

procedure TComplexEdit.SetRe(Value: Extended);
begin
  if Value <> x.Value then
   begin x.Value := Value; FNewValue := True end
end;

procedure TComplexEdit.SetIm(Value: Extended);
begin
  if Value <> y.Value then
   begin y.Value := Value; FNewValue := True end
end;

function TComplexEdit.GetCValue: TComplexRec;
begin
  if FNewValue then
   begin
     FCValue := SetComplex(x.Value, y.Value);
     FNewValue := False; FNoNorm := True
   end;
  Result := FCValue
end;

function TComplexEdit.GetRe: Extended;
begin Result := x.Value end;

function TComplexEdit.GetIm: Extended;
begin Result := y.Value end;

function TComplexEdit.GetNorm: Extended;
begin
  if FNewValue or FNoNorm then
     begin FNorm := Sqrt(GetSqrNorm); FNewNorm := False end
  else if FNewNorm then begin FNorm := Sqrt(FSqrNorm); FNewNorm := False end;
  Result := FNorm
end;

function TComplexEdit.GetSqrNorm: Extended;
begin
  if FNewValue or FNoNorm then
   begin
     FSqrNorm := CSqrNorm(GetCValue); FNoNorm := False; FNewNorm := True
   end;
  Result := FSqrNorm
end;

function TComplexEdit.GetArg: Extended;
begin
  if FNewValue or FNoArg then
     begin FArg := CArg(GetCValue); FNoArg := False end;
  Result := FArg
end;

procedure TComplexEdit.SetNorm(Value: Extended);
var R: Extended;
begin
  if Value <> GetNorm then
   begin
     if y.Value = 0 then y.Value := Value
     else
      begin
        R := Value/FNorm; x.Value := x.Value * R; y.Value := y.Value * R
      end;
     FNorm := Value; FSqrNorm := Sqr(Value)
   end
end;

procedure TComplexEdit.SetArg(Value: Extended);
begin
  if Value <> GetArg then
   begin
     FArg := Value;
     x.Value := GetNorm * Cos(Value); y.Value := FNorm * Sin(Value)
   end
end;

function TComplexEdit.GetValueStr: String;
begin Result := x.Text + ' + ' + y.Text + 'i' end;

procedure TComplexEdit.AssignValues(Source: TComplexEdit);
begin
  x.Value := Source.x.Value; y.Value := Source.y.Value;
  FNewValue := Source.FNewValue;
  if not FNewValue then
   begin
     FCValue := Source.FCValue;
     FNoNorm := Source.FNoNorm;
     if not FNoNorm then
      begin
        FSqrNorm := Source.FSqrNorm;
        FNewNorm := Source.FNewNorm;
        FNorm := Source.FNorm
      end;
     FNoArg := Source.FNoArg;
     FArg := Source.FArg
   end
end;

procedure TComplexEdit.FrameExit(Sender: TObject);
begin
  if not FNewValue then
     FNewValue := (x.Value <> FCValue.x) or (y.Value <> FCValue.y)
end;

procedure TComplexEdit.Inc(const Ix, Iy: Extended);
begin
  SetValues(x.Value + Ix, y.Value + Iy)
end;

procedure TComplexEdit.Inc(const I: TComplexRec);
begin with I do Inc(x, y) end;

procedure TComplexEdit.Dec(const D: TComplexRec);
begin with D do Inc(-x, -y) end;

procedure TComplexEdit.IncRe(const I: Extended);
begin SetRe(x.Value + I) end;

procedure TComplexEdit.IncIm(const I: Extended);
begin SetIm(y.Value + I) end;

procedure TComplexEdit.Scale(const Scal: Extended; KeepNorm: Boolean = False);
begin
  x.Value := x.Value*Scal;
  y.Value := y.Value*Scal;
  if KeepNorm and not (FNewValue or FNoNorm) then
   begin
     FCValue := SetComplex(x.Value, y.Value);
     if not FNewNorm then FNorm := FNorm*Scal;
     FSqrNorm := FSqrNorm*Sqr(Scal)
   end
  else FNewValue := True
end;

end.
