unit ComplTrig;

interface
uses Complex;

function Cexp(const z: TComplexRec): TPolComplRec;
function CDexp(const z: TComplexRec): TComplexRec;
function Cln(const z: TPolComplRec): TComplexRec;
function CDln(const z: TComplexRec): TComplexRec;

function Csin(const z: TComplexRec): TComplexRec;
function Ccos(const z: TComplexRec): TComplexRec;
procedure Trig(const Z: TComplexRec; out CosZ, SinZ: TComplexRec);

function CArctgh(const z: TComplexRec): TComplexRec;
function CArctg(const z: TComplexRec): TComplexRec;

implementation
uses Math;

function Cexp(const z: TComplexRec): TPolComplRec;
begin with z do Result := SetPolar(Exp(x), y) end;

function CDexp(const z: TComplexRec): TComplexRec;
begin Result := CPolarToDec(Cexp(z)) end;

function Cln(const z: TPolComplRec): TComplexRec;
begin with z do Result := SetComplex(Ln(r), a) end;

function CDln(const z: TComplexRec): TComplexRec;
begin CDln := CLn(ComplToPolar(z)) end;

function Csin(const z: TComplexRec): TComplexRec;
begin with z do Result := SetComplex(Sin(x)*Cosh(y), Cos(x)*Sinh(y)) end;

function Ccos(const z: TComplexRec): TComplexRec;
begin with z do Result := SetComplex(Cos(x)*Cosh(y), -Sin(x)*Sinh(y)) end;

procedure Trig(const Z: TComplexRec; out CosZ, SinZ: TComplexRec);
var Ch, Cs, E, Sh, Sn: Extended;
begin
  with Z do
   begin
     Cs := Cos(x); Sn := Sin(x);
     if Abs(y) >= 1E-4
     then begin E := Exp(y); Ch := (E+1/E)/2; Sh := (E-1/E)/2 end
     else begin Ch := Cosh(y); Sh := (y*y/6.0+1)*y end
   end;
  with CosZ do begin x := Cs*Ch; y := -Sn*Sh end;
  with SinZ do begin x := Sn*Ch; y := Sn*Sh end
end;

function CArctgh(const z: TComplexRec): TComplexRec;
begin
 with z do
  Result := CScal(Cln(ComplToPolar(
                         CDiv(SetComplex(1+x,y),SetComplex(1-x,-y)))), 0.5)
end;

function CArctg(const z: TComplexRec): TComplexRec;
var t: TComplexRec;
begin
with z do t := CArctgh(SetComplex(-y,x)); with t do Result := SetComplex(-y,x)
end;

end.
