unit Complex;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, RealEdit, NumStrCt;

type
  TComplexRec = record x, y: Extended end;
  TPolComplRec = record r, a: Extended end;

  TCustomComplex = class(TCustomPanel)
  private
    { Private declarations }
    V1, V2: TRealEdit;
    d: TComplexRec;
    p: TPolComplRec;
    s: Extended;
    ValType: Byte;
    procedure CheckCValue(V: Byte);
    procedure CheckPolarVal(V: Byte);
    procedure ScalNorm(const Scaler, NewNorm: Extended);
  protected
    { Protected declarations }
    function CanResize(var NewWidth: Integer; var NewHeight: Integer):
      Boolean; override;
    procedure EditExit(Sender: TObject); virtual;
    function GetFormatControl: TComponent;
    procedure SetFormatControl(NewValue: TComponent);
    function GetStringFormat1: TNumStringFormat;
    function GetStringFormat2: TNumStringFormat;
    procedure SetStringFormat(NewValue: TNumStringFormat);
    function GetValueFormat: Boolean;
    procedure SetValueFormat(NewValue: Boolean);
    function GetNextOnEnter: TWinControl;
    procedure SetNextOnEnter(NewValue: TWinControl);
    function GetRaiseConvertError: Boolean;
    procedure SetRaiseConvertError(NewValue: Boolean);
    function GetOnConvertError: TNotifyEvent;
    procedure SetOnConvertError(NewValue: TNotifyEvent);
    function GetHideSelection: Boolean;
    procedure SetHideSelection(NewValue: Boolean);
    function GetMaxLength: Integer;
    procedure SetMaxLength(NewValue: Integer);
    function GetModified: Boolean;
    procedure SetModified(NewValue: Boolean);
    function GetReadOnly: Boolean;
    procedure SetReadOnly(NewValue: Boolean);
    function GetOnChange: TNotifyEvent;
    procedure SetOnChange(NewValue: TNotifyEvent);
    function CheckValType: Byte; virtual; abstract;
    procedure CheckValChange; virtual; abstract;
    procedure CheckNewValues; virtual; abstract;
    procedure ShowValues(ViewType: Byte = 0); virtual; abstract;
    function GetCValue: TComplexRec; virtual;
    procedure SetCValue(NewValue: TComplexRec);
    function GetRe: Extended; virtual; abstract;
    procedure SetRe(NewValue: Extended); virtual;
    function GetIm: Extended; virtual; abstract;
    procedure SetIm(NewValue: Extended); virtual;
    function GetPolVal: TPolComplRec; virtual;
    procedure SetPolVal(NewValue: TPolComplRec);
    function GetNorm: Extended; virtual;
    procedure SetNorm(NewValue: Extended); virtual;
    function GetArg: Extended; virtual;
    procedure SetArg(NewValue: Extended); virtual;
    function GetSqrNorm: Extended;
    procedure SetSqrNorm(NewValue: Extended);
    function GetCalcAll: Boolean;
    procedure SetCalcAll(NewValue: Boolean); virtual; abstract;
    function GetText: string; virtual;
    function GetText1: string;
    function GetText2: string;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure AssignValues(Source: TCustomComplex); virtual; abstract;
    function NewValues: Boolean; virtual;
    function MainValues: Boolean; virtual;
    procedure Scale(const Scaler: Extended); virtual; abstract;
    procedure Conj;
    procedure Inv;
    function Inc(const z: TComplexRec): TComplexRec; overload;
    function Inc(const z: TCustomComplex): TComplexRec; overload;
    function Dec(const z: TComplexRec): TComplexRec; overload;
    function Dec(const z: TCustomComplex): TComplexRec; overload;
    function ScalMult(const z: TCustomComplex): Extended;
    procedure Mult(const z: TCustomComplex; ToPolar: Boolean = False);
    function CMult(const z: TComplexRec): TComplexRec; overload;
    function CMult(const z: TCustomComplex): TComplexRec; overload;
    function PolarMult(const z: TPolComplRec): TPolComplRec; overload;
    function PolarMult(const z: TCustomComplex): TPolComplRec; overload;
    property Modifted: Boolean read GetModified write SetModified default False;
    property CValue: TComplexRec read GetCValue write SetCValue;
    property Re: Extended read GetRe write SetRe;
    property Im: Extended read GetIm write SetIm;
    property PolVal: TPolComplRec read GetPolVal write SetPolVal;
    property Norm: Extended read GetNorm write SetNorm;
    property Arg: Extended read GetArg write SetArg;
    property SqrNorm: Extended read GetSqrNorm write SetSqrNorm;
    property CalcAll: Boolean read GetCalcAll write SetCalcAll default False;
    property Text: string read GetText;
    property Text1: string read GetText1;
    property Text2: string read GetText2;
  published
    { Published declarations }
    property FormatControl: TComponent
      read GetFormatControl write SetFormatControl default Nil;
    property StringFormat1: TNumStringFormat
      read GetStringFormat1 write SetStringFormat;
    property StringFormat2: TNumStringFormat
      read GetStringFormat2 write SetStringFormat;
    property ValueFormat: Boolean
      read GetValueFormat write SetValueFormat default False;
    property NextOnEnter: TWinControl
      read GetNextOnEnter write SetNextOnEnter;
    property RaiseConvertError: Boolean
      read GetRaiseConvertError write SetRaiseConvertError default False;
    property OnConvertError: TNotifyEvent
      read GetOnConvertError write SetOnConvertError;
    property HideSelection: Boolean
      read GetHideSelection write SetHideSelection default True;
    property MaxLength: Integer
      read GetMaxLength write SetMaxLength default 0;
    property ReadOnly: Boolean
      read GetReadOnly write SetReadOnly default False;
    property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
    property Anchors;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property Color;
    property Constraints;
    property Ctl3D;
    property Enabled;
    property Font;
    property FullRepaint;
    property HelpContext;
    property ParentCtl3d;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

  TDComplex = class(TCustomComplex)
  private
    { Private declarations }
  protected
    { Protected declarations }
    function CheckValType: Byte; override;
    procedure CheckValChange; override;
    procedure CheckNewValues; override;
    procedure ShowValues(ViewType: Byte = 0); override;
    function GetRe: Extended; override;
    function GetIm: Extended; override;
    procedure SetNorm(NewValue: Extended); override;
    procedure SetArg(NewValue: Extended); override;
    procedure SetCalcAll(NewValue: Boolean); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure AssignValues(Source: TCustomComplex); override;
    procedure Scale(const Scaler: Extended); override;
  published
    { Published declarations }
    property Re;
    property Im;
  end;

  TPolarCompl = class(TCustomComplex)
  private
    { Private declarations }
  protected
    { Protected declarations }
    function CheckValType: Byte; override;
    procedure CheckValChange; override;
    procedure CheckNewValues; override;
    procedure ShowValues(ViewType: Byte = 0); override;
    function GetRe: Extended; override;
    procedure SetRe(NewValue: Extended); override;
    function GetIm: Extended; override;
    procedure SetIm(NewValue: Extended); override;
    function GetPolVal: TPolComplRec; override;
    function GetNorm: Extended; override;
    procedure SetNorm(NewValue: Extended); override;
    function GetArg: Extended; override;
    procedure SetCalcAll(NewValue: Boolean); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure AssignValues(Source: TCustomComplex); override;
    procedure Scale(const Scaler: Extended); override;
  published
    { Published declarations }
    property Norm;
    property Arg;
  end;

  TComplex = class(TCustomComplex)
  private
    { Private declarations }
    FPolarView: Boolean;
  protected
    { Protected declarations }
    procedure DblClick; override;
    function CheckValType: Byte; override;
    procedure CheckValChange; override;
    procedure CheckNewValues; override;
    procedure ShowValues(ViewType: Byte = 0); override;
    function GetCValue: TComplexRec; override;
    function GetRe: Extended; override;
    procedure SetRe(NewValue: Extended); override;
    function GetIm: Extended; override;
    procedure SetIm(NewValue: Extended); override;
    function GetPolVal: TPolComplRec; override;
    function GetNorm: Extended; override;
    procedure SetNorm(NewValue: Extended); override;
    function GetArg: Extended; override;
    procedure SetArg(NewValue: Extended); override;
    procedure SetCalcAll(NewValue: Boolean); override;
    procedure SetPolarView(NewValue: Boolean);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure AssignValues(Source: TCustomComplex); override;
    procedure Scale(const Scaler: Extended); override;
  published
    { Published declarations }
    property PolarView: Boolean
      read FPolarView write SetPolarView default False;
  end;

const
  pi_2 = pi/2;
  CZero: TComplexRec = (x: 0; y: 0);
  Im1: TComplexRec = (x: 0; y: 1);
  Re1: TComplexRec = (x: 1; y: 0);
  PolarZero: TPolComplRec = (r: 0; a: 0);
  Im1polar: TPolComplRec = (r: 1; a: pi_2);
  Re1polar: TPolComplRec = (r: 1; a: 0);

function SetComplex(const Re, Im: Extended): TComplexRec;

function CSqrNorm(const z: TComplexRec): Extended;
function CNorm(const z: TComplexRec): Extended;
function CArg(const z: TComplexRec): Extended;
function ScalMult(const u, v: TComplexRec): Extended;

function CScale(var z: TComplexRec; const Scaler: Extended): TComplexRec;
function CInc(var z: TComplexRec; const y: TComplexRec): TComplexRec;
function CMult(var z: TComplexRec; const y: TComplexRec): TComplexRec;

function CScal(const z: TComplexRec; const Scaler: Extended): TComplexRec;
function CConj(const z: TComplexRec): TComplexRec;
function CSqr(const z: TComplexRec): TComplexRec;
function CSqrt(const z: TComplexRec): TComplexRec;
function CInv(const z: TComplexRec): TComplexRec;
function CDiff(const z, diff: TComplexRec): TComplexRec; overload;
function CDiff(const r: Extended; diff: TComplexRec): TComplexRec; overload;
function CDiv(const z, divid: TComplexRec): TComplexRec;
function CSum(const a, b: TComplexRec): TComplexRec; overload;
function CSum(const a: array of TComplexRec): TComplexRec; overload;
function CSum(const z: TComplexRec; const r: Extended): TComplexRec; overload;
function CProd(const a, b: TComplexRec): TComplexRec; overload;
function CProd(const a: array of TComplexRec): TComplexRec; overload;

function SetPolar(const Norm, Arg: Extended): TPolComplRec;

function PolarScalMult(const u, v: TPolComplRec): Extended;

function PolarScal(const z: TPolComplRec; const Scaler: Extended):
         TPolComplRec;
function PolarConj(const z: TPolComplRec): TPolComplRec;
function PolarMult(var z: TPolComplRec; const y: TPolComplRec): TPolComplRec;
function PolarSquare(const z: TPolComplRec): TPolComplRec;
function PolarInv(const z: TPolComplRec): TPolComplRec;
function PolarDiv(const z, d: TPolComplRec): TPolComplRec;
function PolarProd(const z, u: TPolComplRec): TPolComplRec; overload;
function PolarProd(const a: array of TPolComplRec): TPolComplRec; overload;

function ComplToPolar(const z: TComplexRec): TPolComplRec;
function ComplToPolarSN(const z: TComplexRec; out SNorm: Extended):
         TPolComplRec;
function CPolarToDec(const z: TPolComplRec): TComplexRec;


procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('vaMathEd', [TDComplex, TPolarCompl, TComplex]);
end;


function SetComplex(const Re, Im: Extended): TComplexRec;
begin with Result do begin x := Re; y := Im end end;

function CSqrNorm(const z: TComplexRec): Extended;
begin with z do Result := x*x + y*y end;

function CNorm(const z: TComplexRec): Extended;
begin Result := Sqrt(CSqrNorm(z)) end;

function CArg(const z: TComplexRec): Extended;
begin
 with z do
  if Abs(y) > Abs(x)
  then if y > 0
       then Result := pi_2 - ArcTan(x/y)
       else Result := -pi_2 - ArcTan(x/y)
  else if x = 0 then Result := 0
  else if x > 0 then Result := ArcTan(y/x)
  else if y < 0
       then Result := -pi + ArcTan(y/x)
       else Result := pi + ArcTan(y/x)
end;

function ScalMult(const u, v: TComplexRec): Extended;
begin Result := u.x * v.x + u.y * v.y end;

function CScale(var z: TComplexRec; const Scaler: Extended): TComplexRec;
begin with z do begin x := x*Scaler; y := y*Scaler end; Result := z end;

function CInc(var z: TComplexRec; const y: TComplexRec): TComplexRec;
begin z.x := z.x + y.x; z.y := z.y + y.y; Result := z end;

function CMult(var z: TComplexRec; const y: TComplexRec): TComplexRec;
begin
  z.x := z.x * y.x - z.y * y.y;
  z.y := z.y * y.x + z.x * y.y;
  Result := z
end;

function CScal(const z: TComplexRec; const Scaler: Extended): TComplexRec;
begin with z do Result := SetComplex(x*Scaler, y*Scaler) end;

function CConj(const z: TComplexRec): TComplexRec;
begin with z do Result := SetComplex(x, -y) end;

function CSqr(const z: TComplexRec): TComplexRec;
begin with z do Result := SetComplex(x*x - y*y, x*y*2) end;

function CSqrt(const z: TComplexRec): TComplexRec;
var D: Extended;
begin
  D := CNorm(z);
  with Result do begin x := sqrt((D + z.x)*0.5); y := sqrt((D - z.x)*0.5) end;
  if z.y < 0 then Result.y := -Result.y
end;

function CInv(const z: TComplexRec): TComplexRec;
var s: Extended;
begin
  s := CSqrNorm(z);
  with z do Result := SetComplex(x/s, -y/s)
end;

function CDiff(const z, diff: TComplexRec): TComplexRec;
begin with Result do begin x := z.x - diff.x; y := z.y - diff.y end end;

function CDiff(const r: Extended; diff: TComplexRec): TComplexRec;
begin with Result do begin x := r - diff.x; y := - diff.y end end;

function CDiv(const z, divid: TComplexRec): TComplexRec;
var s: Extended;
begin
  s := CSqrNorm(divid);
  with Result do
   begin
     x := (z.x * divid.x + z.y * divid.y)/s;
     y := (z.y * divid.x - z.x * divid.y)/s
   end
end;

function CSum(const a, b: TComplexRec): TComplexRec;
begin with Result do begin x := a.x + b.x; y := a.y + b.y end end;

function CSum(const a: array of TComplexRec): TComplexRec;
var I: Integer; S: TComplexRec;
begin
  S := a[0]; for I := 1 to High(a) do CInc(S, a[I]); Result := S
end;

function CSum(const z: TComplexRec; const r: Extended): TComplexRec;
begin with Result do begin x := z.x + r; y := z.y end end;

function CProd(const a, b: TComplexRec): TComplexRec;
begin
with Result do begin x := a.x * b.x - a.y * b.y; y := a.x * b.y + a.y * b.x end
end;

function CProd(const a: array of TComplexRec): TComplexRec;
var I: Integer; P: TComplexRec;
begin
  P := a[0]; for I := 1 to High(a) do CMult(P, a[I]); Result := P
end;

function SetPolar(const Norm, Arg: Extended): TPolComplRec;
begin with Result do begin
  r := Abs(Norm);
  if r >= 0 then a := Arg else if Arg >= 0 then a := a + pi else a := a - pi
end end;

function PolarScalMult(const u, v: TPolComplRec): Extended;
begin Result := u.r * v.r * Cos(u.a - v.a) end;

function PolarScal(const z: TPolComplRec; const Scaler: Extended):
         TPolComplRec;
begin with Result do begin
  r := z.r * Scaler;
  if r >= 0 then a := z.a
  else
   begin
     r := Abs(r); if z.a >= 0 then a := z.a - pi else a := z.a + pi
   end
end end;

function PolarConj(const z: TPolComplRec): TPolComplRec;
begin with Result do begin  r := z.r; a := -z.a end end;

function PolarMult(var z: TPolComplRec; const y: TPolComplRec): TPolComplRec;
begin z.r := z.r * y.r; z.a := z.a + y.a; Result := z end;

function PolarSquare(const z: TPolComplRec): TPolComplRec;
begin with Result do begin r := Sqr(z.r); a := z.a*2 end end;

function PolarInv(const z: TPolComplRec): TPolComplRec;
begin with Result do begin r := 1/z.r;  a := -z.a end end;

function PolarDiv(const z, d: TPolComplRec): TPolComplRec;
begin with Result do begin r := z.r / d.r; a := z.a - d.a end end;

function PolarProd(const z, u: TPolComplRec): TPolComplRec;
begin with Result do begin r := z.r * u.r; a := z.a + u.a end end;

function PolarProd(const a: array of TPolComplRec): TPolComplRec;
var I: Integer; P: TPolComplRec;
begin
  P := a[0]; for I := 1 to High(a) do PolarMult(P, a[I]); Result := P
end;

function ComplToPolar(const z: TComplexRec): TPolComplRec;
begin with Result do begin r := CNorm(z); a := CArg(z) end end;

function ComplToPolarSN(const z: TComplexRec; out SNorm: Extended):
         TPolComplRec;
begin with Result do begin
  SNorm := CSqrNorm(z); r := Sqrt(SNorm); a := CArg(z)
end end;

function CPolarToDec(const z: TPolComplRec): TComplexRec;
begin with Result do begin x := z.r * Cos(z.a); y := z.r * Sin(z.a) end end;


constructor TCustomComplex.Create(AOwner: TComponent); {override;}
begin
  inherited;
  Height := 21;
  Width := 153;
  V1 := TRealEdit.Create(Self);
  V1.Parent := Self;
  V1.Width := 64;
  V1.StringFormat.Pressigion := 4;
  V2 := TRealEdit.Create(Self);
  V2.Parent := Self;
  with V2 do
   begin
    Width := 64;
    Left := 89;
    Anchors := [akTop, akRight];
    StringFormat.Pressigion := 4;
   end;
  V1.NextOnEnter := V2;
  V1.OnExit := EditExit;
  V2.OnExit := EditExit;
  ValType := 7
end;

constructor TDComplex.Create(AOwner: TComponent); {override;}
begin
  inherited;
  Caption := ' +_i* '
end;

constructor TPolarCompl.Create(AOwner: TComponent); {override;}
begin
  inherited;
  Caption := ' e^i* '
end;

constructor TComplex.Create(AOwner: TComponent); {override;}
begin
  inherited;
  Caption := ' +_i* '
end;

//procedure TCustomComplex.AssignValues(Source: TCustomComplex);virtual;abstract;

procedure TDComplex.AssignValues(Source: TCustomComplex);{override;}
begin
  if Source is TDComplex
     or ((Source is TComplex) and not TComplex(Source).FPolarView)
  then begin V1.Value := Source.V1.Value; V2.Value := Source.V2.Value end
  else SetCValue(Source.GetCValue);
  d := Source.d; p := Source.p; s := Source.s;
  ValType := Source.ValType
end;

procedure TPolarCompl.AssignValues(Source: TCustomComplex);{override;}
begin
  if Source is TPolarCompl
     or ((Source is TComplex) and TComplex(Source).FPolarView)
  then begin V1.Value := Source.V1.Value; V2.Value := Source.V2.Value end
  else SetPolVal(Source.GetPolVal);
  d := Source.d; p := Source.p; s := Source.s;
  ValType := Source.ValType
end;

procedure TComplex.AssignValues(Source: TCustomComplex);{override;}
begin
  if Source is TComplex then SetPolarView(TComplex(Source).FPolarView)
  else if Source is TDComplex then SetPolarView(False)
  else if Source is TPolarCompl then SetPolarView(True);
  V1.Value := Source.V1.Value; V2.Value := Source.V2.Value;
  d := Source.d; p := Source.p; s := Source.s;
  ValType := Source.ValType
end;

function TCustomComplex.CanResize(var NewWidth, NewHeight: Integer):
  Boolean; {override}
begin
  V1.Width := (NewWidth - Canvas.TextWidth(Caption)) div 2;
  with V1 do
   begin
     if Width < Font.Size then Width := Font.Size;
     if NewHeight < Abs(Font.Height) then NewHeight := Abs(Font.Height);
     Height := NewHeight
   end;
  with V2 do
   begin
     Width := V1.Width;
     Left := NewWidth - Width;
     Height := NewHeight
   end;
  if V2.Left < V1.Width + Canvas.TextWidth(Caption) then
   begin
     V2.Left := V1.Width + Canvas.TextWidth(Caption);
     NewWidth := V2.Left + V2.Width
   end;
  Result := inherited CanResize(NewWidth, NewHeight)
end;

procedure TCustomComplex.EditExit(Sender: TObject); {virtual;}
begin CheckValChange end;

function TCustomComplex.GetFormatControl: TComponent;
begin Result := V1.FormatControl end;

procedure TCustomComplex.SetFormatControl(NewValue: TComponent);
begin
  V1.FormatControl := NewValue;
  V2.FormatControl := NewValue;
end;

function TCustomComplex.GetStringFormat1: TNumStringFormat;
begin Result := V1.StringFormat end;

function TCustomComplex.GetStringFormat2: TNumStringFormat;
begin Result := V2.StringFormat end;

procedure TCustomComplex.SetStringFormat(NewValue: TNumStringFormat);
begin
  V1.StringFormat := NewValue;
  V2.StringFormat := NewValue
end;

function TCustomComplex.GetValueFormat: Boolean;
begin Result := V1.ValueFormat end;

procedure TCustomComplex.SetValueFormat(NewValue: Boolean);
begin
  V1.ValueFormat := NewValue;
  V2.ValueFormat := NewValue
end;

function TCustomComplex.GetNextOnEnter: TWinControl;
begin Result := V2.NextOnEnter end;

procedure TCustomComplex.SetNextOnEnter(NewValue: TWinControl);
begin
  V2.NextOnEnter := NewValue
end;

function TCustomComplex.GetRaiseConvertError: Boolean;
begin Result := V1.RaiseConvertError end;

procedure TCustomComplex.SetRaiseConvertError(NewValue: Boolean);
begin
  V1.RaiseConvertError := NewValue;
  V2.RaiseConvertError := NewValue
end;

function TCustomComplex.GetOnConvertError: TNotifyEvent;
begin Result := V1.OnConvertError end;

procedure TCustomComplex.SetOnConvertError(NewValue: TNotifyEvent);
begin
  V1.OnConvertError := NewValue;
  V2.OnConvertError := NewValue
end;

function TCustomComplex.GetHideSelection: Boolean;
begin Result := V1.HideSelection end;

procedure TCustomComplex.SetHideSelection(NewValue: Boolean);
begin
  V1.HideSelection := NewValue;
  V2.HideSelection := NewValue
end;

function TCustomComplex.GetMaxLength: Integer;
begin Result := V1.MaxLength end;

procedure TCustomComplex.SetMaxLength(NewValue: Integer);
begin
  V1.MaxLength := NewValue;
  V2.MaxLength := NewValue
end;

function TCustomComplex.GetModified: Boolean;
begin Result := V1.Modified or V2.Modified end;

procedure TCustomComplex.SetModified(NewValue: Boolean);
begin
  V1.Modified := NewValue;
  V2.Modified := NewValue
end;

function TCustomComplex.GetReadOnly: Boolean;
begin Result := V1.ReadOnly end;

procedure TCustomComplex.SetReadOnly(NewValue: Boolean);
begin
  V1.ReadOnly := NewValue;
  V2.ReadOnly := NewValue
end;

function TCustomComplex.GetOnChange: TNotifyEvent;
begin Result := V1.OnChange end;

procedure TCustomComplex.SetOnChange(NewValue: TNotifyEvent);
begin
  V1.OnChange := NewValue;
  V2.OnChange := NewValue
end;

procedure TComplex.DblClick;
begin
  CheckValChange;
  SetPolarView(not FPolarView);
  inherited
end;

{ValType:
 2:  d,p.r   corresponds to V1,V2     $20: d,p.r   differs from V1,V2
 3:  d,p.r,s corresponds to V1,V2     $30: d,p.r,s differs from V1,V2
 4:  d       correspond  to V1,V2     $40: d       differ  from V1,V2
 5:  d, s    corresponds to V1,V2     $50: d, s    differs from V1,V2
 6:  d,p     corresponds to V1,V2     $60: d,p     differs from V1,V2
 7:  d,p,s   corresponds to V1,V2     $70: d,p,s   differs from V1,V2
 8:  d,p.a   corresponds to V1,V2     $80: d,p.a   differs from V1,V2
 9:  d,p.a,s corresponds to V1,V2     $90: d,p.a,s differs from V1,V2
 $A: p       correspond  to V1,V2     $A0: p       differs from V1,V2
 $B: p, s    corresponds to V1,V2     $B0: p, s    differs from V1,V2
}

procedure TCustomComplex.CheckCValue(V: Byte);
begin
  if V > $10 then with d do
     begin x := V1.Value; y := V2.Value; ValType := 4 end
end;

procedure TCustomComplex.CheckPolarVal(V: Byte);
begin
  if V > $10 then with p do
   begin
     if Odd(ValType) and (r = V1.Value) then ValType := $B
     else begin r := V1.Value; ValType := $A end;
     a := V2.Value
   end
end;

procedure TCustomComplex.ScalNorm(const Scaler, NewNorm: Extended);
begin
  if ValType <= 9 then CScale(d, Scaler);
  if (ValType mod 4 > 1) then
   begin
     p.r := Abs(NewNorm);
     if (NewNorm < 0) and (ValType in [6..8]) then
        if p.a > 0 then p.a := p.a - pi else p.a := p.a + pi
   end;
  if Scaler <> 1 then
   begin
     if Odd(ValType) then System.Dec(ValType);
     ValType := ValType shl 4
   end
end;

// function TCustomComplex.CheckValType: Byte; virtual; abstact;

function TDComplex.CheckValType: Byte; {override;}
begin
  if (ValType > $10) and (V1.Value = d.x) and (V2.Value = d.y) then
     ValType := ValType shr 4;
  Result := ValType
end;

function TPolarCompl.CheckValType: Byte; {override;}
begin
  if (ValType > $10) and (V1.Value = p.r) and (V2.Value = p.a) then
     ValType := ValType shr 4;
  Result := ValType
end;

function TComplex.CheckValType: Byte; {override;}
begin
  if ValType > $10 then
     if FPolarView
     then
      begin
        if (V1.Value = p.r) and (V2.Value = p.a) then
           ValType := ValType shr 4
      end
     else if (V1.Value = d.x) and (V2.Value = d.y) then
          ValType := ValType shr 4;
  Result := ValType
end;

// procedure TCustomComplex.CheckValChange; virtual; abstract;

procedure TDComplex.CheckValChange; {override;}
begin
 with d do
   if (V1.Value = x) and (V2.Value = y) then
      begin if ValType > $10 then ValType := ValType shr 4 end
   else if ValType < $10 then ValType := ValType shl 4
end;

procedure TPolarCompl.CheckValChange; {override;}
begin
 with p do
   if (V1.Value = r) or (V2.Value = a) then
      begin if ValType > $10 then ValType := ValType shr 4 end
   else if ValType < $10 then ValType := ValType shl 4
end;

procedure TComplex.CheckValChange; {override;}
begin
  if FPolarView then with p do
   begin
     if (V1.Value = r) and (V2.Value = a) then
        begin if ValType > $10 then ValType := ValType shr 4 end
     else if ValType < $10 then ValType := ValType shl 4
   end
  else with d do
     if (V1.Value = x) and (V2.Value = y) then
        begin if ValType > $10 then ValType := ValType shr 4 end
     else if ValType < $10 then ValType := ValType shl 4
end;

// procedure TCustomComplex.CheckNewValues; virtual; abstract;

procedure TDComplex.CheckNewValues; {override;}
begin CheckCValue(CheckValType) end;

procedure TPolarCompl.CheckNewValues; {override;}
begin CheckPolarVal(CheckValType) end;

procedure TComplex.CheckNewValues; {override;}
begin
if FPolarView then CheckPolarVal(CheckValType) else CheckCValue(CheckValType)
end;

// procedure TCustomComplex.ShowValues; virtual; abstract;

procedure TDComplex.ShowValues(ViewType: Byte = 0); {override;}
begin
 if ValType > $10 then
  begin
    ValType := ValType shr 4;
    if ValType > 9 then
       begin d := CPolarToDec(p); System.Dec(ValType, 4) end;
    V1.Value := d.x; V2.Value := d.y
  end
end;

procedure TPolarCompl.ShowValues(ViewType: Byte = 0); {override;}
begin
 if ValType > $10 then
  begin
    ValType := ValType shr 4;
    case ValType of
     2..3: begin p.a := CArg(d); System.Inc(ValType, 4) end;
     4:    begin
             s := CSqrNorm(d); p.r := Sqrt(s); p.a := CArg(d); ValType := 7
           end;
     5:    begin p.r := Sqrt(s); p.a := CArg(d); ValType := 7 end;
     8:    begin s := CSqrNorm(d); p.r := Sqrt(s); ValType := 7 end;
     9:    begin p.r := Sqrt(s); ValType := 7 end;
    end;
    V1.Value := p.r; V2.Value := p.a
  end
end;

procedure TComplex.ShowValues(ViewType: Byte = 0); {override;}
begin
 if ValType > $10 then
  begin
    ValType := ValType shr 4;
    if PolarView
    then if (ViewType = 1) or ((ViewType = 0) and (ValType in [0..5,8..9]))
         then SetPolarView(False)
         else
          begin
            case ValType of
             2..3: begin p.a := CArg(d); System.Inc(ValType, 4) end;
             4:    begin
                     s := CSqrNorm(d); p.r := Sqrt(s); p.a := CArg(d);
                     ValType := 7
                   end;
             5:    begin p.r := Sqrt(s); p.a := CArg(d); ValType := 7 end;
             8:    begin s := CSqrNorm(d); p.r := Sqrt(s); ValType := 7 end;
             9:    begin p.r := Sqrt(s); ValType := 7 end;
            end;
            V1.Value := p.r; V2.Value := p.a
          end
    else if (ViewType = 2) or ((ViewType = 0) and (ValType > 9))
         then SetPolarView(True)
         else
          begin
            if ValType > 9 then
               begin d := CPolarToDec(p); System.Dec(ValType, 4) end;
            V1.Value := d.x; V2.Value := d.y
          end
  end
end;

function TCustomComplex.GetCValue: TComplexRec;{virtual;}
begin
  CheckNewValues;
  if ValType > 9 then begin d := CPolarToDec(p); System.Dec(ValType, 4) end;
  Result := d
end;

// function TDComplex, TPolarCompl.GetCValue = inherited;

function TComplex.GetCValue: TComplexRec;{override;}
begin
  Result := inherited GetCValue;
  SetPolarView(False)
end;

procedure TCustomComplex.SetCValue(NewValue: TComplexRec);
begin
  if (NewValue.x <> d.x) or (NewValue.y <> d.y) then
     begin d := NewValue; ValType := $40 end;
  ShowValues(1)
end;

// function TCustomComplex.GetRe: Extended; virtual ; abstract;

function TDComplex.GetRe: Extended; {override;}
begin Result := V1.Value end;

function TPolarCompl.GetRe: Extended; {override;}
begin Result := GetCValue.x end;

function TComplex.GetRe: Extended; {overrige;}
begin
  if FPolarView then Result := GetCValue.x else Result := V1.Value
end;

procedure TCustomComplex.SetRe(NewValue: Extended); {virtual;}
begin
  if V1.Value <> NewValue then
   begin
     V1.Value := NewValue; if ValType < $10 then ValType := ValType shl 4
   end
end;

// procedure TDComplex.SetRe(NewValue: Extended) = inherited

procedure TPolarCompl.SetRe(NewValue: Extended); {override;}
begin
  if GetRe <> NewValue then SetCValue(SetComplex(NewValue, d.y))
end;

procedure TComplex.SetRe(NewValue: Extended); {override;}
begin
  SetPolarView(False);
  inherited
end;

// function TComplex.GetIm: Extended; virtual ; abstract;

function TDComplex.GetIm: Extended; {override;}
begin Result := V2.Value end;

function TPolarCompl.GetIm: Extended; {override;}
begin Result := GetCValue.y end;

function TComplex.GetIm: Extended; {overrige;}
begin
  if FPolarView then Result := GetCValue.y else Result := V2.Value
end;

procedure TCustomComplex.SetIm(NewValue: Extended); {virtual;}
begin
  if V2.Value <> NewValue then
   begin
     V2.Value := NewValue; if ValType < $10 then ValType := ValType shl 4
   end
end;

// procedure TDComplex.SetIm(NewValue: Extended) = inherited

procedure TPolarCompl.SetIm(NewValue: Extended); {override;}
begin
  if GetIm <> NewValue then SetCValue(SetComplex(d.x, NewValue))
end;

procedure TComplex.SetIm(NewValue: Extended); {override;}
begin
  SetPolarView(False);
  inherited
end;

function TCustomComplex.GetPolVal: TPolComplRec; {virtual;}
begin
  GetNorm;
  if ValType < 6 then begin p.a := CArg(d); System.Inc(ValType, 4) end;
  Result := p
end;

// function TDComplex.GetPolVal: TPolComplRec = inherited

function TPolarCompl.GetPolVal: TPolComplRec; {override;}
begin
  CheckPolarVal(CheckValType);
  Result := p
end;

function TComplex.GetPolVal: TPolComplRec; {override;}
begin
  if FPolarView then CheckPolarVal(CheckValType) else SetPolarView(False);
  Result := p
end;

procedure TCustomComplex.SetPolVal(NewValue: TPolComplRec); {virtual;}
begin
  if NewValue.r <> p.r
  then begin p := NewValue; ValType := $A0 end
  else if NewValue.a <> p.a then
   begin
     p := NewValue;
     if Odd(ValType) or Odd(ValType shr 4)
     then ValType := $B
     else ValType := $A
   end;
  ShowValues(2)
end;

function TCustomComplex.GetNorm: Extended; {virtual;}
begin
  CheckNewValues;
  case ValType of
   4,8: begin
          s := CSqrNorm(d); p.r := Sqrt(s); System.Dec(ValType)
        end;
   5,9: begin p.r := Sqrt(s); System.Dec(ValType, 2) end;
  end;
  Result := p.r
end;

// function TDComplex.GetNorm: Extended; = inherited

function TPolarCompl.GetNorm: Extended; {override;}
begin Result := V1.Value end;

function TComplex.GetNorm: Extended; {override;}
begin
  if FPolarView then Result := V1.Value
  else
   begin
     Result := inherited GetNorm; if ValType >=6 then SetPolarView(True)
   end
end;

procedure TCustomComplex.SetNorm(NewValue: Extended); {virtual;}
begin
  if NewValue <> GetNorm then
     if CheckValType <= 9 then
        if p.r <> 0 then ScalNorm(NewValue/p.r, NewValue)
        else begin p.r := NewValue; d := CPolarToDec(p); ValType := 6 end
     else ValType := $A0
end;

procedure TDComplex.SetNorm(NewValue: Extended); {override;}
begin
  inherited;
  ShowValues
end;

procedure TPolarCompl.SetNorm(NewValue: Extended); {override;}
begin
  inherited;
  V1.Value := Abs(NewValue);
  if NewValue < 0 then
     if V2.Value > 0
     then V2.Value := V2.Value - pi
     else V2.Value := V2.Value + pi
end;

procedure TComplex.SetNorm(NewValue: Extended); {override;}
begin
  inherited;
  if FPolarView then
   begin
     V1.Value := NewValue;
     if NewValue < 0 then
        if V2.Value > 0
        then V2.Value := V2.Value - pi
        else V2.Value := V2.Value + pi
   end
  else ShowValues(Ord(ValType > $50)*2)
end;

function TCustomComplex.GetArg: Extended; {virtual;}
begin
  if not (CheckValType in [6..$B]) then
     begin CheckCValue(ValType); p.a := CArg(d); System.Inc(ValType, 4) end;
  Result := p.a
end;

// function TDComplex.GetArg: Extended; = inherited

function TPolarCompl.GetArg: Extended; {override;}
begin Result := V2.Value end;

function TComplex.GetArg: Extended; {override;}
begin
  if FPolarView then Result := V2.Value
  else
   begin
     Result := inherited GetArg;
     if (ValType mod 4 >= 2) then SetPolarView(True)
   end
end;

procedure TCustomComplex.SetArg(NewValue: Extended); {virtual;}
begin
  if NewValue <> V2.Value then
   begin
     V2.Value := NewValue; if ValType < $10 then ValType := ValType shl 4
   end
end;

procedure TDComplex.SetArg(NewValue: Extended); {override}
begin SetPolVal(SetPolar(GetNorm, NewValue)) end;

// procedure TPolarComplex.SetArg(NewValue: Extended) = inherited

procedure TComplex.SetArg(NewValue: Extended); {override}
begin
  SetPolarView(True);
  inherited
end;

function TCustomComplex.GetSqrNorm: Extended;
begin
  CheckNewValues;
  if not Odd(ValType) then
   begin
     if (ValType mod 4) = 2 then s := Sqr(p.r) else s := CSqrNorm(d);
     System.Inc(ValType)
   end;
  Result := s
end;

procedure TCustomComplex.SetSqrNorm(NewValue: Extended);
begin
  if (NewValue >= 0) and (Odd(CheckValType) or (NewValue <> s)) then
   begin
     SetNorm(Sqrt(NewValue)); s := NewValue;
     if not Odd(ValType) then
        if ValType < $10 then System.Inc(ValType)
        else if not Odd(ValType shr 4) then System.Inc(ValType, $10)
   end
end;

function TCustomComplex.GetCalcAll: Boolean;
begin Result := CheckValType = 7 end;

//procedure TCustomComplex.SetCalcAll(NewValue: Boolean); virtual;abstract;

procedure TDComplex.SetCalcAll(NewValue: Boolean); {override;}
begin
  if NewValue and not GetCalcAll then GetPolVal
end;

procedure TPolarCompl.SetCalcAll(NewValue: Boolean); {override;}
begin
  if NewValue and not GetCalcAll then begin GetCValue; GetSqrNorm end
end;

procedure TComplex.SetCalcAll(NewValue: Boolean); {override;}
begin
  if NewValue and not GetCalcAll then
     if FPolarView
     then begin inherited GetCValue; GetSqrNorm end
     else inherited GetPolVal
end;

procedure TComplex.SetPolarView(NewValue: Boolean);
var W,H: Integer;
begin
  if NewValue <> FPolarView then
   begin
     if NewValue
     then
      begin
        V1.Value := inherited GetPolVal.r; V2.Value := p.a;
        Caption := ' e^i* ';
      end
     else
      begin
        V1.Value := inherited GetCValue.x; V2.Value := d.y;
        Caption := ' +_i* ';
      end;
     W := Width; H := Height; CanResize(W, H);
     if Width < W then Width := W;
     FPolarView := NewValue
   end
end;

function TCustomComplex.GetText: string;
begin
  Result := V1.Text + Caption + V2.Text
end;

function TCustomComplex.GetText1: string;
begin Result := V1.Text end;

function TCustomComplex.GetText2: string;
begin Result := V2.Text end;

function TCustomComplex.NewValues: Boolean;
begin
  Result := CheckValType > $10
end;

function TCustomComplex.MainValues: Boolean;
begin
  Result := CheckValType in [4,$A]
end;

procedure TDComplex.Scale(const Scaler: Extended); {override}
begin
  CheckCValue(CheckValType);
  ScalNorm(Scaler, Abs(p.r)*Scaler);
  with d do begin V1.Value := x; V2.Value := y end
end;

procedure TPolarCompl.Scale(const Scaler: Extended); {override;}
begin
  CheckPolarVal(CheckValType);
  ScalNorm(Scaler, p.r*Scaler);
  with p do begin V1.Value := r; V2.Value := a end
end;

procedure TComplex.Scale(const Scaler: Extended); {override;}
begin
  if FPolarView then with p do
   begin
     CheckPolarVal(CheckValType);
     ScalNorm(Scaler, p.r*Scaler);
     V1.Value := r; V2.Value := a
   end
  else with d do
   begin
     CheckCValue(CheckValType);
     ScalNorm(Scaler, p.r*Scaler);
     V1.Value := x; V2.Value := y
   end
end;

procedure TCustomComplex.Conj;
begin
  d.x := -d.x; p.a := -p.a;
  V2.Value := -V2.Value
end;

procedure TCustomComplex.Inv;
begin
  d.x := d.x/GetSqrNorm; d.y := -d.y/s; p.r := 1/p.r; p.a := -p.a;
  ShowValues
end;

function TCustomComplex.Inc(const z: TComplexRec): TComplexRec;
var t: TComplexRec;
begin
  t := GetCValue; SetCValue(CInc(t, z)); Result := d
end;

function TCustomComplex.Inc(const z: TCustomComplex): TComplexRec;
begin Result := Inc(z.GetCValue) end;

function TCustomComplex.Dec(const z: TComplexRec): TComplexRec;
begin
  SetCValue(CDiff(GetCValue, z))
end;

function TCustomComplex.Dec(const z: TCustomComplex): TComplexRec;
begin Result := Dec(z.GetCValue) end;

function TCustomComplex.ScalMult(const z: TCustomComplex): Extended;
begin
  CheckNewValues;
  case ValType of
   6..7:   begin
             z.CheckNewValues;
             if z.ValType in [$A..$B]
             then Result := Complex.PolarScalMult(p, z.p)
             else Result := Complex.ScalMult(d, z.GetCValue)
           end;
   $A..$B: begin
             z.CheckNewValues;
             if z.ValType in [6..7,$A..$B]
             then Result := Complex.PolarScalMult(p, z.p)
             else Result := Complex.ScalMult(GetCValue, z.d)
           end;
   else           Result := Complex.ScalMult(d, z.GetCValue)
  end
end;

procedure TCustomComplex.Mult(const z: TCustomComplex;
                              ToPolar: Boolean = False); {virtual;}
begin
  CheckNewValues;
  case ValType of
   6..7:   begin
             z.CheckNewValues;
             case z.ValType of
              6..7:   begin
                        Complex.CMult(d, z.d);
                        Complex.PolarMult(p, z.p);
                        ValType := $60
                      end;
              $A..$B: begin
                        Complex.PolarMult(p, z.p);
                        ValType := $A0
                      end;
              else    begin Complex.CMult(d, z.d); ValType := $40 end;
             end
           end;
   $A..$B: begin
             if not ToPolar then with z do
              begin CheckNewValues; ToPolar := ValType in [6..7,$A..$B] end;
             if ToPolar
             then begin Complex.PolarMult(p, z.GetPolVal); ValType := $A0 end
             else
              begin
                d := CPolarToDec(p); Complex.CMult(d, z.d); ValType := $40
              end
           end;
   else    begin
             if ToPolar then with z do
                begin CheckNewValues; ToPolar := ValType > 9 end;
             if ToPolar
             then
              begin
                p := ComplToPolar(d);Complex.PolarMult(p, z.p);ValType :=$A0
              end
             else begin Complex.CMult(d, z.GetCValue); ValType := $40 end
           end
  end;
  ShowValues
end;

function TCustomComplex.CMult(const z: TComplexRec): TComplexRec;
var t: TComplexRec;
begin
  t := GetCValue; SetCValue(Complex.CMult(t, z)); Result := d
end;

function TCustomComplex.CMult(const z: TCustomComplex): TComplexRec;
begin
  Mult(z, False);
  Result := GetCValue
end;

function TCustomComplex.PolarMult(const z: TPolComplRec): TPolComplRec;
var t: TPolComplRec;
begin
  t := GetPolVal; SetPolVal(Complex.PolarMult(t, z)); Result := p
end;

function TCustomComplex.PolarMult(const z: TCustomComplex): TPolComplRec;
begin
  Mult(z, True);
  Result := GetPolVal
end;

end.
