unit Interpol;

interface

type
  TrealFunction = function(x: Extended): Extended;
  TReal2Function = function(x, y: Extended): Extended;
  TTableIdValue = function(Id: Integer; Init, Step: Extended): Extended;
  TValueToTab = function(Value, Init, Step: Extended): Extended;
  TRTableValue = function(P: Extended; Init, Step: Extended): Extended;

function LinScaledId(Id: Integer; Init, Step: Extended): Extended;
function LogScaledId(Id: Integer; Init, Step: Extended): Extended;
function ValToLinTabPos(Value, Init, Step: Extended): Extended;
function ValToLogTabPos(Value, Init, Step: Extended): Extended;
function LinScaledValue(P: Extended; Init, Step: Extended): Extended;
function LogScaledValue(P: Extended; Init, Step: Extended): Extended;

function LinTabDer(Value, Step: Extended): Extended;// dValToLinTabPos/dValue
function LogTabDer(Value, Step: Extended): Extended;// dValToLogTabPos/dValue

function Eq4Interpol(fp, f0, f1, f2, r: Extended): Extended;//r=(x-x0)/(x1-x0),
//    interpolation to 0..1, f(0)=f0, f(1)=f1, f'(0)=(f1-fp)/2; f'(1)=(f2-f0)/2
function InvEq4Interp(fp, f0, f1, f2, fValue: Extended): Extended;
// Result in 0..1 if fValue in f0..f1, Eq4interpol(fp,f0,f1,f2, Result) = fValue
function P4Interpol(xp,fp, x0,f0, x1,f1, x2,f2, x: Extended): Extended;
//    interpolation to x in x0..x1, f(0)=f0, f(1)=f1,
//                                 f'(0)=(f1-fp)/(x1-xp); f'(1)=(f2-f0)/(x2-xp)
function InvP4Interp(xp,fp, x0,f0, x1,f1, x2,f2, fValue: Extended): Extended;
//Result in x0..x1, Eq4interpol(xp,fp, x0,f0, x1,f1, x2,f2,  Result) = fValue
function P4ClipInterp(xp,fp, x0,f0, x1,f1, x2,f2, x: Extended): Extended;
// function P4Interpol clipped at [f0..f1]
function GrEq3Interpol(fgr, f1, f2, r: Extended): Extended;//r=(x-xgr)/(x1-xgr),
// extrapolation r <= 1, f(0)=fgr, f(1)=f1, f'(1)=(f2-f0)/2, f'(-inf)=0 or const
function InvGrEq3Interp(fgr, f1, f2, fValue: Extended): Extended;
//                     Result in 0..1, GrEq3etrapol(fgr,f1,f2, Result) = fValue
function GrP3Interpol(xgr,fgr, x1,f1, x2,f2, x: Extended): Extended;
//interpolation  to x in xgr..x1, f(xgr)=fgr, f(1)=f1, f'(1)=(f2-f0)/2
function InvGrP3Interp(xgr,fgr, x1,f1, x2,f2, fValue: Extended): Extended;
//Result in xgr..x1, GrEq3Interpol(fgr,f1,f2, Result) = fValue
function GrP3ClipInterp(xgr,fgr, x1,f1, x2,f2, x: Extended): Extended;
// function GrP3Interpol clipped at [fgr..f1]
function EqLinExtrapol(f0, f1, r: Extended): Extended;//r=(x-x0)/(x1-x0),
// f(x0)=f0, f(x1)=f1, linear extrapolation
function LinExtrapol(x0,f0, x1,f1, x: Extended): Extended;

function dEq4Interp(fp, f0, f1, f2, r: Extended): Extended;
//       dEq4Interpol(fp, f0, f1, f2, r)/dr
function dGrEq3Interp(fgr, f1, f2, r: Extended): Extended;
//       dGrEq3Extrapol(fp, f0, f1, f2, r)/dr
function dP4Interp(xp,fp, x0,f0, x1,f1, x2,f2, x: Extended): Extended;
//       dEq4Interpol(fp, f0, f1, f2, r)/dr
function dGrP3Interp(xgr,fgr, x1,f1, x2,f2, x: Extended): Extended;
//       dGrEq3Extrapol(fp, f0, f1, f2, r)/dr


function InvertFn(f, df: TrealFunction; fValue, Init: Extended): Extended;
//                f(Result) = fValue (df(x) = f'(x), Init - init value of x
function InvertXatY(f, fx: TReal2Function; fValue, xInit, y: Extended):Extended;
//       f(Result, y) = fValue, fx(x,y)=df(x,y)/dx, xInit- init value of x
function InvertYatX(f, fy: TReal2Function; fValue, x, yInit: Extended):Extended;
//       f(x, Result) = fValue, fy(x,y)=df(x,y)/dy, yInit- init value of y

implementation

uses Math, SysUtils;

function LinScaledId(Id: Integer; Init, Step: Extended): Extended;
begin Result := Id * Step + Init end;

function LogScaledId(Id: Integer; Init, Step: Extended): Extended;
begin Result := Exp(Step * Id) * Init end;

function ValToLinTabPos(Value, Init, Step: Extended): Extended;
begin Result := (Value - Init)/Step end;

function ValToLogTabPos(Value, Init, Step: Extended): Extended;
begin Result := ln(Value/Init)/Step end;

function LinScaledValue(P: Extended; Init, Step: Extended): Extended;
begin Result := P * Step + Init end;

function LogScaledValue(P: Extended; Init, Step: Extended): Extended;
begin Result := Exp(Step * P) * Init end;

function LinTabDer(Value, Step: Extended): Extended;
begin Result := 1.0/Step end;

function LogTabDer(Value, Step: Extended): Extended;
begin Result := 1.0/Step/Value end;

function Eq4Interpol(fp, f0, f1, f2, r: Extended): Extended;
var a0, a1, p, q: Extended;
begin
  if f0 = f1 then Result := f0
  else
   begin
     if f0 = fp then a0 := f0 - f1 else a0 := f0 - (f1 + fp)*0.5;
     if f1 = f2 then a1 := f1 - f0 else a1 := f1 - (f0 + f2)*0.5;
     p := 1.0 - r; q := r * p;
     Result := (f0 + a0*q) * p + (f1 + a1*q) * r
   end
end;

function InvEq4Interp(fp, f0, f1, f2, fValue: Extended): Extended;
var a0, a1, a3, d, p, q, r, x1, x2, Press: Extended; Med: Boolean;
begin
  if f0 = f1 then Result := 0.5
  else
   begin
     d := f1 - f0;
     if f0 = fp then a0 := -d else a0 := f0 - (f1 + fp)*0.5;
     if f1 = f2 then a1 :=  d else a1 := f1 - (f0 + f2)*0.5;
     a3 := (a1 - a0)*3.0;
     r := (fValue - f0)/d; x1 := r; Med := False; Press := 1E-20;
     repeat
       if Med then begin Med := False; x1 := (x1 + r)*0.5 end
       else
        begin
          p := 1.0 - x1; q := x1*p;
          x2 := (fValue - (f0+a0*q)*p - (f1+a1*q)*x1)/(d + a0*p-a1*x1+a3*q) +x1;
          if (Abs(x2 - x1) + Abs(x2 - r) <= Press)
          then begin Result := x2; Exit end
          else if x2 = r then begin Med := True; Press := Press * 2.0 end
          else begin r := x1; x1 := x2; Press := Press * 1.2 end
        end
     until x1 = r;
     Result := x1
   end
end;

function P4Interpol(xp,fp, x0,f0, x1,f1, x2,f2, x: Extended): Extended;
var a0, a1, d, q: Extended;
begin
  if f0 = f1 then Result := f0
  else
   begin
     d := x1 - x0; a0 := f1 - f0;
     if f0 = fp then a1 := -a0 else a1 := ((d/(x0 - xp))*(f0-fp)-a0)*d/(x1-xp);
     if f1 <> f2 then a0 := (a0 - (d/(x2 - x1))*(f2-f1))*d/(x2-x0);
     q := (x1 - x)/d; d := (x - x0)/d;
     Result := (f0 + a0*d*d) * q + (f1 + a1*q*q) * d
   end
end;

function P4ClipInterp(xp,fp, x0,f0, x1,f1, x2,f2, x: Extended): Extended;
begin
  if f1 > f0 then
     Result := Min(Max( P4Interpol(xp,fp, x0,f0, x1,f1, x2,f2, x),
                        LinExtrapol(x0,f0, x1,(f0+f1)/2.0, x) ),
                   LinExtrapol(x0,(f0+f1)/2.0, x1,f1, x) )
  else if f1 < f0 then
     Result := Max(Min( P4Interpol(xp,fp, x0,f0, x1,f1, x2,f2, x),
                        LinExtrapol(x0,f0, x1,(f0+f1)/2.0, x) ),
                   LinExtrapol(x0,(f0+f1)/2.0, x1,f1, x) )
  else Result := f0
end;

function InvP4Interp(xp,fp, x0,f0, x1,f1, x2,f2, fValue: Extended): Extended;
var a0, a1, a2, d, p, q, r, t, ri, rii, Press: Extended; Med: Boolean;
begin
  if (f0 = f1) and (fValue = f0) then Result := (x0 + x1)*0.5
  else
   begin
     d := x1 - x0; p := f1 - f0;
     if f0 = fp then a1 := -p else a1 := ((d/(x0 - xp))*(f0-fp)-p)*d/(x1-xp);
     if f1 = f2 then a0 := p else a0 := (p - (d/(x2 - x1))*(f2-f1))*d/(x2-x0);
     a2 := (a0 - a1)*2.0;
     r := (fValue - f0)/p; ri := r; Med := False; Press := 1E-20;
     repeat
       if Med then begin Med := False; ri := (ri + r)*0.5 end
       else
        begin
          t := 1.0 - ri; //q := ri * t;
          rii := ri + (fValue - (f0+a0*ri*ri)*t-(f1+a1*t*t)*ri)/
                      (p - a0*ri*ri+a1*t*t+a2*ri*t);
          if (Abs(rii - ri) + Abs(rii - r) <= Press)
          then begin Result := x0 + rii * d; Exit end
          else if rii = r then begin Med := True; Press := Press * 2.0 end
          else begin r := ri; ri := rii; Press := Press * 1.2 end
        end
     until ri = r;
     Result := x0 + ri * d
   end
end;

function GrEq3Interpol(fgr, f1, f2, r: Extended): Extended;
var a, p, q: Extended;
begin
  if fgr = f1 then Result := fgr
  else
   begin
     if f1 = f2 then a := f1 - fgr else a := f1 - (fgr + f2)*0.5;
     p := 1.0 - r; q := r * p;
     Result := fgr * p + (f1 + a*q) * r
   end
end;

function InvGrEq3Interp(fgr, f1, f2, fValue: Extended): Extended;
var a, a3, d, p, q, r, x1, x2, Press: Extended; Med: Boolean;
begin
  if fgr = f1 then Result := 0
  else
   begin
     d := f1 - fgr;
     if f1 = f2 then a :=  d else a := f1 - (fgr + f2)*0.5;
     a3 := a*3.0;
     r := (fValue - fgr)/d; x1 := r; Med := False; Press := 1E-20;
     repeat
       if Med then begin Med := False; x1 := (x1 + r)*0.5 end
       else
        begin
          p := 1.0 - x1; q := x1*p;
          x2 := (fValue - fgr*p - (f1+a*q)*x1)/(d - a*x1 + a3*q) + x1;
          if (Abs(x2 - x1) + Abs(x2 - r) <= Press)
          then begin Result := x2; Exit end
          else if x2 = r then begin Med := True; Press := Press * 2.0 end
          else begin r := x1; x1 := x2; Press := Press * 1.2 end
        end
     until x1 = r;
     Result := x1
   end
end;

function GrP3Interpol(xgr,fgr, x1,f1, x2,f2, x: Extended): Extended;
var a, d, p: Extended;
begin
  if fgr = f1 then Result := fgr
  else
   begin
     d := x1 - xgr; p := fgr - f1;
     if f1 = f2 then a := p else a := ((f2-f1)*d/(x2-x1)+p)*d/(x2-xgr);
     p := (x - x1)/d; d := (x - xgr)/d;
     Result := -fgr * p + (f1 + a*p*d) * d
   end
end;

function GrP3ClipInterp(xgr,fgr, x1,f1, x2,f2, x: Extended): Extended;
begin
  if f1 > fgr then
     Result := Min(Max( GrP3Interpol(xgr,fgr, x1,f1, x2,f2, x),
                        LinExtrapol(xgr,fgr, x1,(fgr+f1)/2.0, x) ),
                   Min( LinExtrapol(xgr,(fgr+f1)/2.0, x1,f1, x),
                        LinExtrapol(xgr,fgr, x1,f1, x) ) )
  else if f1 < fgr then
     Result := Max(Min( GrP3Interpol(xgr,fgr, x1,f1, x2,f2, x),
                        LinExtrapol(xgr,fgr, x1,(fgr+f1)/2.0, x) ),
                   Max( LinExtrapol(xgr,(fgr+f1)/2.0, x1,f1, x),
                        LinExtrapol(xgr,fgr, x1,f1, x) ) )
  else Result := fgr
end;

function InvGrP3Interp(xgr,fgr, x1,f1, x2,f2, fValue: Extended): Extended;
var a, a3, d, p, q, t, r, ri, rii, Press: Extended; Med: Boolean;
begin
  if fgr = f1 then Result := xgr
  else
   begin
     d := x1 - xgr; p := fgr - f1;
     if f1 = f2 then a := p else a := ((f2-f1)*d/(x2-x1) + p)*d/(x2-xgr);
     a3 := a*3.0;
     r := (fgr - fValue)/p; ri := r; Med := False; Press := 1E-20;
     repeat
       if Med then begin Med := False; ri := (ri + r)*0.5 end
       else
        begin
          t := ri - 1.0; q := ri * t;
          rii := ri - (fValue + fgr*t-(f1+a*q)*ri)/(p - a*ri - a3*q);
          if (Abs(rii - ri) + Abs(rii - r) <= Press)
          then begin Result := xgr + rii * d; Exit end
          else if rii = r then begin Med := True; Press := Press * 2.0 end
          else begin r := ri; ri := rii; Press := Press * 1.2 end
        end
     until ri = r;
     Result := xgr + ri * d
   end
end;

function EqLinExtrapol(f0, f1, r: Extended): Extended;
begin Result := f0*(1.0-r) + f1*r end;

function LinExtrapol(x0,f0, x1,f1, x: Extended): Extended;
begin
  Result := (f0*(x1 - x) + f1*(x - x0))/(x1 - x0)
end;

function dEq4Interp(fp, f0, f1, f2, r: Extended): Extended;
var a0, a1, p: Extended;
begin
  if f0 = f1 then Result := 0
  else
   begin
     if f0 = fp then a0 := f0 - f1 else a0 := f0 - (f1 + fp)*0.5;
     if f1 = f2 then a1 := f1 - f0 else a1 := f1 - (f0 + f2)*0.5;
     p := 1.0 - r;
     Result :=  f1 - f0 + a0*p - a1*r - (a0-a1)*r*p*3.0
   end
end;

function dGrEq3Interp(fgr, f1, f2, r: Extended): Extended;
var a, p: Extended;
begin
  if fgr = f1 then Result := 0
  else
   begin
     if f1 = f2 then a := f1 - fgr else a := f1 - (fgr + f2)*0.5;
     p := 1.0 - r;
     Result :=  f1 - fgr - a*r + a*r*p*3.0
   end
end;

function dP4Interp(xp,fp, x0,f0, x1,f1, x2,f2, x: Extended): Extended;
var a0, a1, d, p, q, r: Extended;
begin
  if f0 = f1 then Result := 0
  else
   begin
     d := x1 - x0; p := (f1 - f0);
     if f0 = fp then a0 := p else a0 := (p - (f0-fp)*d/(x0-xp))*d/(x1-xp);
     if f1 = f2 then a1 := -p else a1 := ((f2-f1)*d/(x2-x1) - p)*d/(x2-x0);
     q := (x - x1)/d; r := (x - x0)/d;
     Result := (p + a0*q + a1*r + 3.0*(a0+a1)*q*r)/d
   end
end;

function dGrP3Interp(xgr,fgr, x1,f1, x2,f2, x: Extended): Extended;
var a, d, p, q, r: Extended;
begin
  if fgr = f1 then Result := 0
  else
   begin
     d := x1 - xgr; p := (fgr - f1);
     if f1 = f2 then a := p else a := ((f2-f1)*d/(x2-x1) + p)*d/(x2-xgr);
     q := (x - x1)/d; r := (x - xgr)/d;
     Result := (-p + a*r + 3.0*a*q*r)/d
   end
end;


function InvertFn(f, df: TrealFunction; fValue, Init: Extended): Extended;
var x1, x2, Press: Extended; Med: Boolean;
begin
  x1 := Init; Med := False; Press := 1E-20;
  repeat
    if Med then begin Med := False; x1 := (x1 + Init)*0.5 end
    else
     try
       x2 := x1 + (fValue - f(x1))/df(x1);
       if (Abs(x2 - x1) + Abs(x2 - Init) <= Abs(x2)*Press)
       then begin Result := x2; Exit end
       else if x2 = Init then begin Med := True; Press := Press * 2.0 end
       else begin Init := x1; x1 := x2; Press := Press * 1.2 end
     except on EMathError do if x1 = Init then raise else Med := True
     end
  until x1 = Init;
  Result := x1
end;

function InvertXatY(f, fx: TReal2Function; fValue, xInit, y: Extended):Extended;
var x1, x2, Press: Extended; Med: Boolean;
begin
  x1 := xInit; Med := False; Press := 1E-20;
  repeat
    if Med then begin Med := False; x1 := (x1 + xInit)*0.5 end
    else
     try
       x2 := x1 + (fValue - f(x1, y))/fx(x1, y);
       if (Abs(x2 - x1) + Abs(x2 - xInit) <= Abs(x2)*Press)
       then begin Result := x2; Exit end
       else if x2 = xInit then begin Med := True; Press := Press * 2.0 end
       else begin xInit := x1; x1 := x2; Press := Press * 1.2 end
     except on EMathError do if x1 = xInit then raise else Med := True
     end
  until x1 = xInit;
  Result := x1
end;

function InvertYatX(f, fy: TReal2Function; fValue, x, yInit: Extended):Extended;
var y1, y2, Press: Extended; Med: Boolean;
begin
  y1 := yInit; Med := False; Press := 1E-20;
  repeat
    if Med then begin Med := False; y1 := (y1 + yInit)*0.5 end
    else
     try
       y2 := y1 + (fValue - f(x, y1))/fy(x, y1);
       if (Abs(y2 - y1) + Abs(y2 - yInit) <= Abs(y2)*Press) then
          begin Result := y2; Exit end
       else if y2 = yInit then
          begin Med := True; Press := Press * 2.0 end
       else
          begin yInit := y1; y1 := y2; Press := Press * 1.2 end

     except on EMathError do if y1 = yInit then raise else Med := True
     end
  until y1 = yInit;
  Result := y1
end;


end.
