unit Lin3diag;
{  3-   
  B[0]*Y[0] + C[0]*Y[1] = F[0],
A[I]*Y[I-1] + B[I]*Y[I] + C[I]*Y[I+1] = F[I], I = 1..N-1,
A[N]*Y[N-1] + B[N]*Y[N] = F[N];
      
Y[0] = Y0 / Y[N] = Yq,
Y'[0] = 0 (Y[0] = Y[1]) / Y'[N] = 0 (Y[N] = Y[N-1])
   Bound0Equation  Bound0Equation}

interface

type
  ABCFrecord = record A, B, C, F: Extended end;
  uvrecord = record u, v: Extended end;
  TBound0Equation = procedure(var Z: uvrecord; B, C, F: Extended);
  TBoundQEquation = function(const Z: uvrecord; A, B, F: Extended): Extended;

var
  Bound0Equation: TBound0Equation;
  BoundQEquation: TBoundQEquation;
  Value0, ValueQ: Extended;

procedure ZeroA0(var Z: uvrecord; B, C, F: Extended);
procedure Keep0Value(var Z: uvrecord; B, C, F: Extended);
procedure ZeroGrad0(var Z: uvrecord; B, C, F: Extended);

function ZeroCn(const Z: uvrecord; A, B, F: Extended): Extended;
function KeepQValue(const Z: uvrecord; A, B, F: Extended): Extended;
function ZeroGradQ(const Z: uvrecord; A, B, F: Extended): Extended;

procedure Gauss(const A, B, C, F: array of Extended;// 
                var Y: array of Extended;// 
                N: Integer = 0//   - 1,  )
                ); overload;

procedure Gauss(const L: array of ABCFrecord;// 
                var Y: array of Extended;// 
                N: Integer = 0//   - 1,  )
                ); overload;

procedure SetBoundaryEquations(KeepY0, KeepYq: Boolean;
                               Y0: Extended = 0; Yq: Extended = 0);

implementation

procedure ZeroA0(var Z: uvrecord; B, C, F: Extended);
begin Z.u := -C/B; Z.v := F/B end;

procedure Keep0Value(var Z: uvrecord; B, C, F: Extended);
begin Z.u := 0; Z.v := Value0 end;

procedure ZeroGrad0(var Z: uvrecord; B, C, F: Extended);
begin Z.u := -2.0*C/B; Z.v := F/B end;

function ZeroCn(const Z: uvrecord; A, B, F: Extended): Extended;
begin Result := (F - A*Z.v)/(A*Z.u + B) end;

function KeepQValue(const Z: uvrecord; A, B, F: Extended): Extended;
begin Result := ValueQ end;

function ZeroGradQ(const Z: uvrecord; A, B, F: Extended): Extended;
begin Result := (F - 2.0*A*Z.v)/(2.0*A*Z.u + B) end;

procedure Gauss(const A, B, C, F: array of Extended;// 
                var Y: array of Extended;// 
                N: Integer = 0//   - 1,  
                );
var I: Integer; Z: array of uvrecord;
begin
  if N = 0 then N := High(Y);
  SetLength(Z, N);
  Bound0Equation(Z[0], B[0], C[0], F[0]);
  for I := 1 to N-1 do with Z[I-1] do
   begin
     Z[I].u := -C[I]/(A[I]*u + B[I]);
     Z[I].v := (F[I] - A[I]*v)/(A[I]*u + B[I])
   end;
  Y[N] := BoundQEquation(Z[N-1], A[N], B[N], F[N]);
  for I := N-1 downto 0 do with Z[I] do Y[I] := Y[I+1]*u + v;
  Finalize(Z)
end;

procedure Gauss(const L: array of ABCFrecord;// 
                var Y: array of Extended;// 
                N: Integer = 0//   - 1,  
                );
var I: Integer; Z: array of uvrecord;
begin
  if N = 0 then N := High(Y);
  SetLength(Z, N);
  with L[0] do Bound0Equation(Z[0], B, C, F);
  for I := 1 to N-1 do with L[I], Z[I-1] do
   begin
     Z[I].u := -C/(A*u + B);
     Z[I].v := (F - A*v)/(A*u + B)
   end;
  with L[N] do Y[N] := BoundQEquation(Z[N-1], A, B, F);
  for I := N-1 downto 0 do with Z[I] do Y[I] := Y[I+1]*u + v;
  Finalize(Z)
end;

procedure SetBoundaryEquations(KeepY0, KeepYq: Boolean;
                               Y0: Extended = 0; Yq: Extended = 0);
begin
  if KeepY0 then
   begin Bound0Equation := Keep0Value; Value0 := Y0 end
  else Bound0Equation := ZeroA0;
  if KeepYq then
   begin BoundQEquation := KeepQValue; Valueq := Yq end
  else BoundQEquation := ZeroCn
end;

end.
