unit NumStrCt;

interface

uses
  SysUtils, Classes;

const
  nfsDigits = ['0'..'9'];
  nfsSign = ['-','+'];
  nfsDigitAll = nfsSign + nfsDigits;
  nfsExponent = ['D','E','Q','d','e','q'];
  nfsDefault = nfsDigitAll + nfsExponent;

type
  TPressigionRange = -1..18;

  TNumStringFormat = class(TPersistent)
  private
    { Private declarations - unit only}
    FFormatType: TFloatFormat;//ffNumber for First-Point (Fortran) format
    FPressigion: TPressigionRange;
    FWidth: ShortInt;
    FFormatString: string;
    FDecSeparator: Char;
    FDefaultDec, NonDefault, FRaiseConvertError: Boolean;
    FDecimals, FDelimiters, FFloatChars, FCharsAllowd: set of Char;
  protected
    { Protected declarations - TNumStringFormat and descedants }
    procedure AssignTo(Dest: TPersistent); override;
    procedure SetFormatString;
    procedure SetFormatType(NewValue: TFloatFormat);
    procedure SetPressigion(NewValue: TPressigionRange);
    procedure SetWidth(NewValue: ShortInt);
    procedure SetDecSeparator(NewValue: Char);
    procedure SetDefaultDec(NewValue: Boolean);
    function GetDecSeparator: Char;
    function CheckChars(var NumStr: String): Boolean;
  public
    { Public declarations - all uses }
    property FormatString: String read FFormatString;
    constructor Create;
    procedure Assign(Source: TPersistent); override;
    function NoCharAllowd(var Key: Char; Text: String;
                          SelPos: Integer): Boolean;
//  String representation of Value according to FormatType, Pressigion and Width
    function NumString(Value: Extended): string;
//  Number record Text formated according to FormatType, Pressigion and Width.
//  If FormatAll = False then formatiog incorreac stiong only
    function FormatText(Text: String; FormatAll: Boolean): string; overload;
    function FormatText(Text: string; var Value: Extended;
      FormatAll: Boolean): string; overload;
//  Check if Text is correct number record
    function CheckString(Text: string): Boolean; overload;
    function CheckString(Text: string; var Value: Extended): Boolean;
      overload;
//  Check if Text is correct number record
//  and formating Text according to FormatType, Pressigion and Width
//  if FormatAll = True or Result = False
    function CheckFormat(var Text: string): Boolean; overload;
    function CheckFormat(var Text: string; FormatAll: Boolean): Boolean;
      overload;
    function CheckFormat(var Text: string; var Value: Extended): Boolean;
      overload;
    function CheckFormat(var Text: string; var Value: Extended;
      FormatAll: Boolean): Boolean; overload;
//  Convert string Text to real Value. If FormatAll then reformat
    function NumStringValue(var Text: string; InValue: Extended = 0;
      FormatAll: Boolean = False): Extended;
//  Search for number record in string Text;
//  MaxLen - maximal length of number record,
//  MaxLen = -1 ->  MaxLen = Width; MaxLen = 0 - unlimited
    function FindNumberStr(const Text: string; out Value: Extended;
             MaxLen: Integer = -1; Reformat: Boolean = False): string;
    function FindIntegerStr(const Text: string; out Value: Integer;
             MaxLen: Integer = -1): string;
    function FindNumber(const Text: string; MaxLen: Integer = -1): Extended;
    function FindInteger(const Text: string; MaxLen: Integer = -1): Integer;
//  Search for number record in string Text_Rest.
//  After procedure Text_Rest - rest of input string after extracted number
    function ExtractNumber(var Text_Rest: string;
             MaxLen: Integer = -1): Extended;
    function ExtractInteger(var Text_Rest: string;
             MaxLen: Integer = -1): Integer;
//  Reads TextStr from file FText if TextStr = '' and extracts number
    function FReadValue(var FText: TextFile; var TextStr: string;
             MaxLen: Integer = -1): Extended;
    function FReadInteger(var FText: TextFile; var TextStr: string;
             MaxLen: Integer = -1): Integer;
  published
    { Published declarations - uses in design time}
    property FormatType: TFloatFormat
      read FFormatType write SetFormatType default ffGeneral;
    property Pressigion: TPressigionRange
      read FPressigion write SetPressigion default -1;
    property Width: ShortInt read FWidth write SetWidth default 0;
    property DecSeparator: Char
      read GetDecSeparator write SetDecSeparator default '.';
    property DefaultDec: Boolean read FDefaultDec write SetDefaultDec;
    property RaiseConvertError: Boolean
      read FRaiseConvertError write FRaiseConvertError default False;
  end;

  TNumStrCt = class(TComponent)
  private
    { Private declarations - unit only }
    FFormatControl: TComponent;
    FNumStringFormat: TNumStringFormat;
  protected
    { Protected declarations - TNumStrCt and descendants }
    procedure AssignTo(Dest: TPersistent); override;
    procedure SetFormatControl(NewValue: TComponent);
    procedure SetNumStringFormat(NewValue: TNumStringFormat);
    procedure SetFormatType(NewValue: TFloatFormat);
    procedure SetPressigion(NewValue: TPressigionRange);
    procedure SetWidth(NewValue: ShortInt);
    procedure SetDecSeparator(NewValue: Char);
    procedure SetDefaultDec(NewValue: Boolean);
    procedure SetRaiseConvertError(NewValue: Boolean);
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    function GetFormatType: TFloatFormat;
    function GetPressigion: TPressigionRange;
    function GetWidth: ShortInt;
    function GetDecSeparator: Char;
    function GetDefaultDec: Boolean;
    function GetRaiseConvertError: Boolean;
  public
    { Public declarations - all uses }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure AssignFormatControl(AControl: TComponent;
      StringFormat: TNumStringFormat);
    property NumStringFormat: TNumStringFormat
      read FNumStringFormat write SetNumStringFormat;
  published
    { Published declarations - uses in design time}
    property FormatControl: TComponent
      read FFormatControl write SetFormatControl default Nil;
    property FormatType: TFloatFormat
      read GetFormatType write SetFormatType default ffGeneral;
    property Pressigion: TPressigionRange
      read GetPressigion write SetPressigion default -1;
    property Width: ShortInt read GetWidth write SetWidth default 0;
    property DecSeparator: Char
      read GetDecSeparator write SetDecSeparator default '.';
    property DefaultDec: Boolean read GetDefaultDec write SetDefaultDec;
    property RaiseConvertError: Boolean
      read GetRaiseConvertError write SetRaiseConvertError default False;
  end;

procedure Register;

implementation

uses Math;

procedure Register;
begin
  RegisterComponents('vaMathEd', [TNumStrCt]);
end;

//   TNumStringFormat

//  { Protected declarations - TNumStringFormat and descendants }

procedure TNumStringFormat.AssignTo(Dest: TPersistent);
begin
  if Dest is TNumStringFormat
  then
   begin
     TNumStringFormat(Dest).FFormatType := FFormatType;
     TNumStringFormat(Dest).FPressigion := FPressigion;
     TNumStringFormat(Dest).FWidth := FWidth;
     TNumStringFormat(Dest).SetFormatString;
     if FDefaultDec
     then TNumStringFormat(Dest).FDefaultDec := True
     else TNumStringFormat(Dest).SetDecSeparator(FDecSeparator)
   end
  else inherited
end;

procedure TNumStringFormat.SetFormatString;
begin
  FFormatString := '%';
  if FWidth <> 0 then FFormatString := FFormatString + IntToStr(FWidth);
  if FPressigion >= 0 then
     FFormatString := FFormatString + '.' + IntToStr(FPressigion);
  case FFormatType of
   ffGeneral:  FFormatString := FFormatString + 'g';
   ffExponent: FFormatString := FFormatString + 'e';
   ffFixed:    FFormatString := FFormatString + 'f';
   ffNumber:   FFormatString := FFormatString + 'e';
   ffCurrency: FFormatString := FFormatString + 'm';
  end
end;

procedure TNumStringFormat.SetFormatType(NewValue: TFloatFormat);
begin
  if (NewValue <> FFormatType) and (NewValue <> ffCurrency) then
   begin
     FFormatType := NewValue;
     case FFormatType of
      ffGeneral:  FFormatString[Length(FFormatString)] := 'g';
      ffExponent: FFormatString[Length(FFormatString)] := 'e';
      ffFixed:    FFormatString[Length(FFormatString)] := 'f';
      ffNumber:   FFormatString[Length(FFormatString)] := 'e';
     end
   end
end;

procedure TNumStringFormat.SetPressigion(NewValue: TPressigionRange);
begin
  if NewValue <> FPressigion then
     begin FPressigion := NewValue; SetFormatString end
end;

procedure TNumStringFormat.SetWidth(NewValue: ShortInt);
begin
  if NewValue <> FWidth then
     begin FWidth := NewValue; SetFormatString end
end;

procedure TNumStringFormat.SetDecSeparator(NewValue: Char);
begin
  if FDefaultDec or (NewValue <> FDecSeparator) then
   begin
     if NewValue in FDecimals then FDecSeparator := NewValue
     else if NewValue in FCharsAllowd then Exit
     else
      begin
        Include(FDecimals, NewValue);
        Include(FFloatChars, NewValue);
        Include(FCharsAllowd, NewValue);
        FDecSeparator := NewValue
      end;
     NonDefault := FDecSeparator <> DecimalSeparator;
     FDefaultDec := False
   end
end;

procedure TNumStringFormat.SetDefaultDec(NewValue: Boolean);
begin
  if NewValue <> FDefaultDec then
   begin
     FDefaultDec := NewValue;
     if FDefaultDec
     then begin NonDefault := False; FDecSeparator := DecimalSeparator end
     else NonDefault := FDecSeparator <> DecimalSeparator
   end
end;

function TNumStringFormat.GetDecSeparator: Char;
begin
  if NonDefault then Result := FDecSeparator else Result := DecimalSeparator
end;

function TNumStringFormat.CheckChars(var NumStr: String): Boolean;
var P: Integer; Decimal, Exponent: Boolean;
begin
  Decimal := False; Exponent := False; Result := True;
  NumStr := Trim(NumStr);
  for P := Length(NumStr) downto 1 do
   if NumStr[P] in FFloatChars
   then
    case NumStr[P] of
     '0'..'9':;
     '-','+': if (P > 1) and not (NumStr[P-1] in (nfsExponent+['*','^']))
              then if NumStr[P-1] in FDelimiters then Delete(NumStr, P-1, 1)
                   else begin Result := False; Exit end;
     else
      if NumStr[P] in nfsExponent
      then
       begin
         NumStr[P] := 'E';
         if Decimal or Exponent then begin Result := False; Exit end
         else case P of
               1: NumStr := '1' + NumStr;
               2: if NumStr[1] in nfsSign then Insert('1', NumStr, 2)
              end;
         Exponent := True
       end
      else if Decimal or ((NumStr[P] = '_') and (FDecSeparator <> '_'))
       then if NumStr[P] in FDelimiters then Delete(NumStr, P, 1)
            else begin Result := False; Exit end
       else if Exponent and (NumStr[P] in FDelimiters) and
               ((Pos('.', Copy(NumStr, 1, P-1)) <> 0) or
               (Pos(DecimalSeparator, Copy(NumStr, 1, P-1)) <> 0) or
               (Pos(FDecSeparator, Copy(NumStr, 1, P-1)) <> 0))
            then Delete(NumStr, P, 1)
            else
             begin
               NumStr[P] := DecimalSeparator;
               case P of
                1: NumStr := '0' + NumStr;
                2: if NumStr[1] in nfsSign then Insert('0', NumStr, 2);
               end;
               Decimal := True
             end
    end
   else if (NumStr[P] in FDelimiters) then Delete(NumStr, P, 1)
   else if (NumStr[P] = '*')
        and ((P = 4) or ((P > 4) and (NumStr[P-4] in (FDelimiters+['*']))))
        and (Copy(NumStr, P-3, 4) = '10**')
   then if P = 4
        then begin Delete(NumStr, P-3, 4); Insert(' 1E', NumStr, 1) end
        else begin Delete(NumStr, P-4, 5); Insert('   E', NumStr, P-4) end
   else if (NumStr[P] = '^')
        and ((P = 3) or ((P > 3) and (NumStr[P-3] in (FDelimiters+['*']))))
        and (Copy(NumStr, P-2, 3) = '10^')
   then if P = 3
        then begin Delete(NumStr, P-2, 3); Insert('1E', NumStr, 1) end
        else begin Delete(NumStr, P-3, 4); Insert('  E', NumStr, P-3) end
   else begin Result := False; Exit end;
  if NumStr = '' then Result := False
end;

//  { Public declarations - all uses of TNumStringFormat }

constructor TNumStringFormat.Create;
begin
  inherited Create;
  FPressigion := -1; FFormatString := '%g';
  FDecSeparator := '.'; FDefaultDec := True;
  FDecimals := ['.',',','_'] + [DecimalSeparator];
  FFloatChars := nfsDefault + FDecimals;
  FDelimiters := [' ',',','_','{','}'] + [ThousandSeparator];
  FCharsAllowd := FFloatChars + FDelimiters
end;

procedure TNumStringFormat.Assign(Source: TPersistent);
begin
  if Source is TNumStringFormat
  then
   begin
     FFormatType := TNumStringFormat(Source).FFormatType;
     FPressigion := TNumStringFormat(Source).FPressigion;
     FWidth := TNumStringFormat(Source).FWidth;
     SetFormatString;
     if TNumStringFormat(Source).FDefaultDec
     then FDefaultDec := True
     else SetDecSeparator(TNumStringFormat(Source).FDecSeparator)
   end
  else inherited
end;

function TNumStringFormat.NoCharAllowd(var Key: Char; Text: String;
         SelPos: Integer): Boolean;
begin
  if Key in FFloatChars then
   case Key of
    '0'..'9': Result := False;
    '-','+':  Result := not ((Trim(Copy(Text, 1, SelPos)) = '')
                             or (Text[SelPos] in nfsExponent));
    else
     if (SelPos = 0) or (Text[SelPos] in nfsDigits)
        or ((SelPos = 1) and (Text[1] in nfsSign))
     then
      begin
        Insert(Key, Text, SelPos+1);
        if CheckChars(Text) then
         begin
           if Key in nfsExponent then Key := 'E' else Key := DecSeparator;
           Result := False
         end
        else Result := True
      end
     else Result := True
   end
  else if Key in FDelimiters
  then Result := not ((Trim(Copy(Text, 1, SelPos)) = '')
                      or (Text[SelPos] in nfsDigits))
  else Result := True
end;

function TNumStringFormat.NumString(Value: Extended): String;
var m, e: string; P: Integer;
begin
  Result := Format(FFormatString, [Value]);
  if NonDefault then
     if Pos(DecimalSeparator, Result) <> 0 then
        Result[Pos(DecimalSeparator, Result)] := FDecSeparator;
  if FFormatType = ffNumber then
  begin
    e := Result; P := Pos('E', Result);
    if P > 0 then
     begin
       m := Trim(Copy(e, 1, P-1)); Delete(e, 1, P); P := StrToInt(Trim(e));
       if m[1] in ['0'..'9'] then m := ' ' + m;
       if m[2] = '0' then begin Delete(m, 2, 1); m := m + '0' end
       else begin Delete(m, 3, 1); Insert('.', m, 2); Inc(P) end;
       e := IntToStr(Abs(P));
       if P >= 0
       then if Length(e) = 1 then m := m + 'E+0' + e else m := m + 'E+' + e
       else if Length(e) = 1 then m := m + 'E-0' + e else m := m + 'E-' + e;
       while Length(m) < Abs(FWidth) do
          if FWidth > 0 then m := ' ' + m else m := m + ' ';
       Result := m
     end
  end
end;

function TNumStringFormat.FormatText(Text: string;
         FormatAll: Boolean): string;
var Check: Boolean; V: Extended;
begin
  if not FormatAll then Result := Text;
  Check := CheckChars(Text);
  try
    V := StrToFloat(Text);
    if FormatAll or not Check then Result := NumString(V)
   except on EConvertError do
     begin
       Check := True;
       Result := '';
       if FRaiseConvertError then raise
     end
  end;
  if (not Check) and FRaiseConvertError then
     raise EConvertError.Create('Not Float Number Char(s)')
end;

function TNumStringFormat.FormatText(Text: string;
  var Value: Extended; FormatAll: Boolean): string;
var S: String; Check: Boolean;
begin
  S := NumString(Value);
  if Trim(Text) = Trim(S)
  then if FormatAll then Result := S else Result := Text
  else
   begin
     if not FormatAll then Result := Text;
     Check := CheckChars(Text);
     try
       Value := StrToFloat(Text);
       if FormatAll or not Check then Result := NumString(Value);
      except on EConvertError do
       begin
         Check := True;
         Result := S;
         if FRaiseConvertError then raise
       end
     end;
     if (not Check) and FRaiseConvertError then
        raise EConvertError.Create('Not Float Number Char(s)')
   end
end;

function TNumStringFormat.CheckString(Text: string): Boolean;
begin
  Result := CheckChars(Text);
  if Result then
     try StrToFloat(Text)
      except on EConvertError do Result := False
     end
end;

function TNumStringFormat.CheckString(Text: string;
  var Value: Extended): Boolean;
begin
  Text := Trim(Text);
  if Text = Trim(NumString(Value))
  then Result := True
  else
   begin
     Result := CheckChars(Text);
     if Result then
        try Value := StrToFloat(Text)
         except on EConvertError do Result := False
        end
   end
end;

function TNumStringFormat.CheckFormat(var Text: String): Boolean;
var S: String;
begin
  S := Text;
  Result := CheckChars(S);
  if Result
  then try
         StrToFloat(S)
        except on EConvertError do begin Text := ''; Result := False end
       end
  else Text := ''
end;

function TNumStringFormat.CheckFormat(var Text: String; FormatAll: Boolean):
  Boolean;
var S: String; V: Extended;
begin
  S := Text;
  Result := CheckChars(S);
  if Result
  then try
         V := StrToFloat(S);
         if FormatAll then Text := NumString(V)
        except on EConvertError do begin Text := ''; Result := False end
       end
  else Text := ''
end;

function TNumStringFormat.CheckFormat(var Text: string; var Value: Extended):
  Boolean;
var T, S: String;
begin
  T := Trim(Text); S := NumString(Value);
  if T = Trim(S)
  then Result := True
  else
   begin
     Result := CheckChars(T);
     if Result
     then try
            Value := StrToFloat(T);
           except on EConvertError do begin Text := S; Result := False end
          end
     else Text := S
   end
end;

function TNumStringFormat.CheckFormat(var Text: string; var Value: Extended;
  FormatAll: Boolean): Boolean;
var T, S: String;
begin
  T := Trim(Text); S := NumString(Value);
  if T = Trim(S)
  then begin Result := True; if FormatAll then Text := S end
  else
   begin
     Result := CheckChars(T);
     if Result
     then try
            Value := StrToFloat(T);
            if FormatAll then Text := NumString(Value)
           except on EConvertError do begin Text := S; Result := False end
          end
     else Text := S
   end
end;

function TNumStringFormat.NumStringValue(var Text: String;
  InValue: Extended; FormatAll: Boolean): Extended;
var T, S: String; Check: Boolean;
begin
  T := Trim(Text); S := NumString(InValue);
  if T = Trim(S)
  then begin if FormatAll then Text := S; Result := InValue end
  else
   begin
     Check := CheckChars(T);
     try
       Result := StrToFloat(T);
       if FormatAll or not Check then Text := NumString(Result)
      except on EConvertError do
       begin
         Text := S; Result := InValue; Check := True;
         if FRaiseConvertError then raise
       end
     end;
     if (not Check) and FRaiseConvertError then
        raise EConvertError.Create('Not Float Number Char(s)')
   end
end;

function TNumStringFormat.FindNumberStr(const Text: string; out Value: Extended;
         MaxLen: Integer = -1; Reformat: Boolean = False): string;
var P, Q: Integer; S: String;
begin
  P := 0;
  repeat
    if Length(Text) > P then Inc(P)
    else begin Value := 0.0; Result := ''; Exit end
  until Text[P] in nfsDigits;
  Q := P;
  while (Q > 1) and (Text[Q-1] in FFloatChars) and (P-Q <= 2) do Dec(Q);
  while Q < P do
   begin
     S := Copy(Text, Q, P - Q + 1); if CheckChars(S) then Break else Inc(Q)
   end;
  if MaxLen < 0 then MaxLen := Max(FWidth, 0);
  if MaxLen = 0 then MaxLen := Length(Text)
                else MaxLen := Min(MaxLen+Q-1, Length(Text));
  while (MaxLen > P) and (Text[P+1] in (FFloatChars+['*','^','_','{','}']))
     do Inc(P);
  S := Copy(Text, Q, P-Q+1);
  Value := 0;
  while not CheckString(S, Value) do Delete(S, Length(S), 1);
  if Reformat then Result := NumString(Value) else Result := S
end;

function TNumStringFormat.FindIntegerStr(const Text: string; out Value: Integer;
         MaxLen: Integer = -1): string;
var P, Q: Integer;
begin
  P := 0;
  repeat
    if Length(Text) > P then Inc(P)
    else begin Value := 0; Result := ''; Exit end
  until Text[P] in nfsDigits;
  if MaxLen < 0 then MaxLen := Max(FWidth, 0);
  if MaxLen = 0 then MaxLen := Length(Text)
                else MaxLen := Min(MaxLen+P-1, Length(Text));
  if (P > 1) and (Text[P-1] in nfsSign) then begin Dec(P); Inc(MaxLen) end;
  Q := P+1;
  while (MaxLen >= Q) and (Text[Q] in nfsDigits) do Inc(Q);
  Result := Copy(Text, P, Q-P);
  Value := StrToInt(Result)
end;

function TNumStringFormat.FindNumber(const Text: string;
         MaxLen: Integer = -1): Extended;
begin FindNumberStr(Text, Result, MaxLen) end;

function TNumStringFormat.FindInteger(const Text: string;
         MaxLen: Integer = -1): Integer;
begin FindIntegerStr(Text, Result, MaxLen) end;

function TNumStringFormat.ExtractNumber(var Text_Rest: string;
         MaxLen: Integer = -1): Extended;
var S: String;
begin
  S := FindNumberStr(Text_Rest, Result, MaxLen);
  if S <> '' then Delete(Text_Rest, 1, Pos(S, Text_Rest) + Length(S) - 1)
end;

function TNumStringFormat.ExtractInteger(var Text_Rest: string;
         MaxLen: Integer = -1): Integer;
var S: String;
begin
  S := FindIntegerStr(Text_Rest, Result, MaxLen);
  if S <> '' then Delete(Text_Rest, 1, Pos(S, Text_Rest) + Length(S) - 1)
end;

function TNumStringFormat.FReadValue(var FText: TextFile; var TextStr: string;
         MaxLen: Integer = -1): Extended;
begin
  while (Trim(TextStr) = '') and not EOF(FText) do ReadLn(FText, TextStr);
  Result := ExtractNumber(TextStr, MaxLen)
end;

function TNumStringFormat.FReadInteger(var FText: TextFile; var TextStr: string;
         MaxLen: Integer = -1): Integer;
begin
  while (Trim(TextStr) = '') and not EOF(FText) do ReadLn(FText, TextStr);
  Result := ExtractInteger(TextStr, MaxLen)
end;

//  TNumStrCt

//  { Protected declarations - TNumStrCt and descendants }

procedure TNumStrCt.AssignTo;
begin
  if Dest is TNumStrCt
  then
    if Assigned(FFormatControl)
    then TNumStrCt(Dest).AssignFormatControl(FFormatControl, FNumStringFormat)
    else TNumStrCt(Dest).AssignFormatControl(Self, FNumStringFormat)
  else if Dest is TNumStringFormat
  then TNumStringFormat(Dest).Assign(FNumStringFormat)
  else inherited
end;

procedure TNumStrCt.SetFormatControl;
var N: TNumStringFormat;
begin
 if NewValue <> FFormatControl then
  if Assigned(NewValue) and (NewValue <> Self)
  then if NewValue is TNumStrCt
       then
         if Assigned(TNumStrCt(NewValue).FFormatControl)
         then AssignFormatControl(TNumStrCt(NewValue).FFormatControl,
                                  TNumStrCt(NewValue).FNumStringFormat)
         else AssignFormatControl(NewValue,
                                  TNumStrCt(NewValue).FNumStringFormat)
       else
        try
          Assign(NewValue)
        except ;
        end
  else if Assigned(FFormatControl) and (FFormatControl <> Self)
  then
   begin
     N := TNumStringFormat.Create; N.Assign(FNumStringFormat);
     FNumStringFormat := N;
     FFormatControl := Nil
   end
end;

procedure TNumStrCt.SetNumStringFormat;
begin
  if (NewValue <> FNumStringFormat) and Assigned(NewValue) then
   begin
     if FNumStringFormat = Nil then FNumStringFormat := TNumStringFormat.Create;
     FNumStringFormat.Assign(NewValue)
   end
end;

procedure TNumStrCt.SetFormatType;
begin FNumStringFormat.SetFormatType(NewValue) end;

procedure TNumStrCt.SetPressigion;
begin FNumStringFormat.SetPressigion(NewValue) end;

procedure TNumStrCt.SetWidth;
begin FNumStringFormat.SetWidth(NewValue) end;

procedure TNumStrCt.SetDecSeparator;
begin FNumStringFormat.SetDecSeparator(NewValue) end;

procedure TNumStrCt.SetDefaultDec;
begin FNumStringFormat.SetDefaultDec(NewValue) end;

procedure TNumStrCt.SetRaiseConvertError;
begin FNumStringFormat.FRaiseConvertError := NewValue end;

procedure TNumStrCt.Notification;
begin
  if (AComponent = FFormatControl) and (Operation = opRemove) then
     SetFormatControl(Nil);
  inherited;
end;

function TNumStrCt.GetFormatType;
begin Result := FNumStringFormat.FFormatType end;

function TNumStrCt.GetPressigion;
begin Result := FNumStringFormat.FPressigion end;

function TNumStrCt.GetWidth;
begin Result := FNumStringFormat.FWidth end;

function TNumStrCt.GetDecSeparator;
begin Result := FNumStringFormat.GetDecSeparator end;

function TNumStrCt.GetDefaultDec;
begin Result := FNumStringFormat.FDefaultDec end;

function TNumStrCt.GetRaiseConvertError;
begin Result := FNumStringFormat.FRaiseConvertError end;

//  { Public declarations - all uses of TNumStrCt }

constructor TNumStrCt.Create;
begin
  inherited;
  FNumStringFormat := TNumStringFormat.Create
end;

destructor TNumStrCt.Destroy;
begin
  if not Assigned(FFormatControl) then FNumStringFormat.Free;
  inherited
end;

procedure TNumStrCt.Assign;
begin
  if Source is TNumStrCt
  then
    if Assigned(TNumStrCt(Source).FFormatControl)
    then AssignFormatControl(TNumStrCt(Source).FFormatControl,
                             TNumStrCt(Source).FNumStringFormat)
    else AssignFormatControl(TNumStrCt(Source),
                             TNumStrCt(Source).FNumStringFormat)
  else if Source is TNumStringFormat
  then
   begin
     if FNumStringFormat = Nil then
        FNumStringFormat := TNumStringFormat.Create;
     FNumStringFormat.Assign(Source)
   end
  else inherited
end;

procedure TNumStrCt.AssignFormatControl;
begin
  if Assigned(AControl) and Assigned(StringFormat) then
   begin
     if (AControl is TNumStrCt) and Assigned(TNumStrCt(AControl).FFormatControl)
     then FFormatControl := TNumStrCt(AControl).FFormatControl
     else FFormatControl := AControl;
     FNumStringFormat := StringFormat
   end
end;

end.
