unit RealEdit;

interface

uses
  SysUtils, Classes, StdCtrls, Controls, NumStrCt;

type
  TRealEdit = class(TCustomEdit)
  private
    { Private declarations }
    FValue, InitValue: Extended;
    FFormatControl: TComponent;
    FStringFormat: TNumStringFormat;
    FRaiseConvertError: Boolean;
    FOnConvertError: TNotifyEvent;
    FOnEnterPress: TNotifyEvent;
    FValueFormat: Boolean;
    FNextOnEnter: TWinControl;
    CheckText: Boolean;
  protected
    { Protected declarations }
    procedure SetValue(NewValue: Extended);
    function GetValue: Extended;
    procedure SetFormatControl(NewValue: TComponent);
    procedure SetStringFormat(NewValue: TNumStringFormat);
    procedure SetValueFormat(NewValue: Boolean);
    procedure SetRaiseConvertError(NewValue: Boolean);
    procedure SetOnConvertError(NewValue: TNotifyEvent);
    procedure AssignTo(Dest: TPersistent); override;
    procedure Change; override;
    procedure DoEnter; override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure KeyPress(var Key: Char); override;
  public
    { Public declarations }
    property Modified;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
  published
    { Published declarations }
    property Value: Extended read GetValue write SetValue;
    property FormatControl: TComponent
      read FFormatControl write SetFormatControl default Nil;
    property StringFormat: TNumStringFormat
      read FStringFormat write SetStringFormat;
    property ValueFormat: Boolean
      read FValueFormat write SetValueFormat default False;
    property RaiseConvertError: Boolean
      read FRaiseConvertError write SetRaiseConvertError default False;
    property NextOnEnter: TWinControl read FNextOnEnter write FNextOnEnter;
    property OnConvertError: TNotifyEvent
      read FOnConvertError write SetOnConvertError;
    property OnEnterPress: TNotifyEvent
      read FOnEnterPress write FOnEnterPress;
    property Anchors;
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property HelpContext;
    property HideSelection;
    property MaxLength;
    property ParentCtl3d;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('vaMathEd', [TRealEdit]);
end;

// TRealEdit - Protected declarations

procedure TRealEdit.SetValue;
begin
  FValue := NewValue;
  Text := FStringFormat.NumString(NewValue);
  CheckText := False
end;

function TRealEdit.GetValue;
begin
  if CheckText then
   begin
     try
       Text := FStringFormat.FormatText(Text, FValue, FValueFormat);
       CheckText := False
      except on EConvertError do
        if Assigned(FOnConvertError)
        then FOnConvertError(Self)
        else if FRaiseConvertError then raise
      end
   end;
  Result := FValue
end;

procedure TRealEdit.SetFormatControl;
var NewFormat: TNumStringFormat; AControl: TNumStrCt;
begin
 if NewValue <> FFormatControl then
  if Assigned(NewValue) and (NewValue <> Self)
  then
   begin
     if NewValue is TNumStrCt
     then AControl := TNumStrCt(NewValue)
     else
      try
        AControl := TNumStrCt.Create(Self);
        AControl.Assign(NewValue)
      except
        AControl.Free;
        Exit
      end;
     if FFormatControl = Nil then FStringFormat.Free;
     if Assigned(AControl.FormatControl)
     then FFormatControl := AControl.FormatControl
     else FFormatControl := AControl;
     FStringFormat := AControl.NumStringFormat;
     if AControl <> NewValue then Acontrol.Free;
     Text := FStringFormat.NumString(FValue);
     CheckText := False
   end
  else if Assigned(FFormatControl) and (FFormatControl <> Self)
  then
   begin
     NewFormat := TNumStringFormat.Create; NewFormat.Assign(FStringFormat);
     FStringFormat := NewFormat;
     FFormatControl := Nil
   end
end;

procedure TRealEdit.SetStringFormat;
begin
  if NewValue <> FStringFormat then
   begin
     FStringFormat.Assign(NewValue);
     Text := FStringFormat.NumString(FValue);
     CheckText := False
   end
end;

procedure TRealEdit.SetValueFormat;
begin
  if NewValue <> FValueFormat then
   begin
     FValueFormat := NewValue;
     if FValueFormat then
      begin
        Text := FStringFormat.NumString(FValue);
        CheckText := False
      end
   end
end;

procedure TRealEdit.SetRaiseConvertError;
begin
  if NewValue <> FRaiseConvertError then
   begin
     FRaiseConvertError := NewValue;
     if FRaiseConvertError then
      begin
        FStringFormat.RaiseConvertError := True;
        FOnConvertError := Nil
      end
   end
end;

procedure TRealEdit.SetOnConvertError;
begin
  FOnConvertError := NewValue;
  if Assigned(FOnConvertError) then
   begin
     FStringFormat.RaiseConvertError := True;
     FRaiseConvertError := False
   end
end;

procedure TRealEdit.AssignTo;
begin
  if Dest is TRealEdit then
   begin
     TRealEdit(Dest).FValue := FValue;
     TRealEdit(Dest).InitValue := InitValue;
     TRealEdit(Dest).SetFormatControl(FFormatControl);
     if (FFormatControl = Nil) and Assigned(FStringFormat) then
        TRealEdit(Dest).SetStringFormat(FStringFormat);
     TRealEdit(Dest).CheckText := CheckText;
     TRealEdit(Dest).FRaiseConvertError := FRaiseConvertError;
     TRealEdit(Dest).FOnConvertError := FOnConvertError;
     TRealEdit(Dest).FOnEnterPress := FOnEnterPress
   end
  else if Dest is TNumStrCt then
   begin
     if Assigned(FFormatControl)
     then TNumStrCt(Dest).AssignFormatControl(FFormatControl, FStringFormat)
     else TNumStrCt(Dest).AssignFormatControl(Self, FStringFormat);
     Exit
   end;
  inherited
end;

procedure TRealEdit.Change;
begin
  inherited;
  CheckText := True
end;

procedure TRealEdit.DoEnter;
begin
  inherited;
  if CheckText and FStringFormat.CheckString(Text, FValue) then
   begin
     CheckText := False;
     InitValue := FValue
   end
end;

procedure TRealEdit.Notification;
begin
  if (AComponent = FFormatControl) and (Operation = opRemove) then
     SetFormatControl(Nil);
  inherited;
end;

procedure TRealEdit.KeyPress;
begin
  if Key < ' '
  then case Key of
        #13: if CheckText and not FStringFormat.CheckString(Text, FValue)
             then SetValue(FValue)
             else
              begin
                if FValueFormat then
                   Text := FStringFormat.NumString(FValue);
                CheckText := False; InitValue := FValue;
                if Assigned(FOnEnterPress) then
                   begin Key := #0; FOnEnterPress(Self) end;
                if Assigned(FNextOnEnter) then
                   begin Key := #0; FNextOnEnter.SetFocus end
              end;
        #27: SetValue(InitValue)
       end
  else if FStringFormat.NoCharAllowd(Key, Text, SelStart)
  then begin Beep; Key := #0 end
end;

// TRealEdit - Public declarations

constructor TRealEdit.Create;
begin
  inherited;
  FStringFormat := TNumStringFormat.Create;
  CharCase := ecUpperCase;
end;

destructor TRealEdit.Destroy;
begin
  if FFormatControl = Nil then FStringFormat.Free;
  inherited
end;

procedure TRealEdit.Assign;
begin
  if Source is TRealEdit then
   begin
     inherited;
     FValue := TRealEdit(Source).FValue;
     if Assigned(TRealEdit(Source).FFormatControl)
     then
      begin
        if FFormatControl = Nil then FStringFormat.Free;
        FFormatControl := TRealEdit(Source).FFormatControl;
        FStringFormat := TRealEdit(Source).FStringFormat
      end
     else
      begin
        if Assigned(TRealEdit(Source).FStringFormat) then
         begin
           if Assigned(FFormatControl) or (FStringFormat = Nil)
           then FStringFormat := TNumStringFormat.Create;
           FStringFormat.Assign(TRealEdit(Source).FStringFormat)
         end;
        FFormatControl := Nil
      end;
     CheckText := TRealEdit(Source).CheckText
   end
  else inherited
end;

end.
