unit Tables;

interface

uses Grids, NumStrCt;

type
  TRealArray = array of Extended;
  TReal2Array = array of array of Extended;
  TIntArray = array of Integer;
  TInt2Array = array of array of Integer;

function FReadTabData(const FileName: String;
         var Values: TReal2Array;
         var Table: TStringGrid;
         ANumFormat: TNumStringFormat = Nil;
         Xdim: Integer = 0;// if < 0 then full, if > 0 then add to FixedColls
         Ydim: Integer = 0;// if < 0 then full, if > 0 then add to FixedRows
         RecLen: Integer = 0): Boolean;
function FReadIntTab(const FileName: String;
         var Values: TInt2Array;
         var Table: TStringGrid;
         NumFormat: TNumStringFormat;
         Xdim, Ydim: Integer; RecLen: Integer = 0): Boolean;

function GetValue(const Data: array of Extended; X: Extended): Extended; overload;
function ValueInCol(const Data: TReal2Array; Col: Integer; Y: Extended):
         Extended; overload;
function ValueInCol(const Data: TReal2Array; const Ydat: array of Extended;
         Col: Integer; Y: Extended; var Yid: Integer): Extended; overload;
function ValueInRow(const Data: TReal2Array; X: Extended; Row: Integer):
         Extended; overload;
function ValueInRow(const Data: TReal2Array; const Xdat: array of Extended;
         X: Extended; Row: Integer; var Xid: Integer): Extended; overload;
function GetValue(const Data: TReal2Array; X, Y: Extended): Extended; overload;
function GetValue(const Data: TReal2Array; const Xdat, Ydat: array of Extended;
         X, Y: Extended; var Xid, Yid: Integer): Extended; overload;

procedure GetPos(const Data: array of Extended; Value: Extended;
          var Ipos: Integer);
function FindValInCol(const Data: TReal2Array; Col: Integer; Value: Extended;
         InitRow: Integer = 0): Extended; overload;
function FindValInCol(const Data: TReal2Array; const Ydat: array of Extended;
         Col: Integer; Value: Extended; var Row: Integer): Extended;overload;
function FindValInRow(const Data: TReal2Array; Row: Integer; Value: Extended;
         InitCol: Integer = 0): Extended; overload;
function FindValInRow(const Data: TReal2Array; const Xdat: array of Extended;
         Row: Integer; Value: Extended; var Col: Integer): Extended;overload;
function FindValXpos(const Data: TReal2Array; Y, Value: Extended): Extended;
         overload;
function FindValXpos(const Data: TReal2Array;
         const Xdat, Ydat: array of Extended;
         Y, Value: Extended; var Xtab, Ytab: Integer): Extended; overload;
function FindValYpos(const Data: TReal2Array; X, Value: Extended): Extended;
         overload;
function FindValYpos(const Data: TReal2Array;
         const Xdat, Ydat: array of Extended;
         X, Value: Extended; var Xtab, Ytab: Integer): Extended; overload;

function Xder(const Data: TReal2Array; X, Y: Extended): Extended; overload;
//       dGetValue/dX
function Xder(const Data: TReal2Array; const Xdat, Ydat: array of Extended;
         X, Y: Extended; var Xid, Yid: Integer): Extended; overload;
function Yder(const Data: TReal2Array; X, Y: Extended): Extended; overload;
//       dGetValue/dY
function Yder(const Data: TReal2Array; const Xdat, Ydat: array of Extended;
         X, Y: Extended; var Xid, Yid: Integer): Extended; overload;

implementation

uses SysUtils, Math, Dialogs, Interpol;

var AddStrNum: Integer = 0;
    AddStr: array of String;
    AddStrRow: set of Byte = [];

function FReadTabData(const FileName: String;
         var Values: TReal2Array;
         var Table: TStringGrid;
         ANumFormat: TNumStringFormat = Nil;
         Xdim: Integer = 0; Ydim: Integer = 0; RecLen: Integer = 0): Boolean;
var X, Xm, Y, Ym, Z, DX, DY, L: Integer;
    S, T: String;
    Numbers: Boolean;
    F: TextFile;
    NumFormat: TNumStringFormat;
function GetFileSize: Integer;
var R: TSearchRec;
 begin
   FindFirst(FileName, faAnyFile, R);
   Result := R.Size
 end;
function FindLabel(var Start: Integer;
                   CollNum: Integer = 0; AskAdd: Boolean = True): String;
var L, Q, N: Integer; T: String; V: Extended;
 begin
   if RecLen = 0 then
    begin
      L := Length(S);
      if S[Start] in [#9,','] then
       begin
         Inc(Start); while S[Start] = ' ' do Inc(Start);
         Result := ''; Exit
       end;
      while (Start <= L) and (S[Start] <= ' ') do Inc(Start);
      Q := Start;
      if Q <= L then
       repeat
         case S[Q] of
          #9, ' ': Break;
          ',': if (Q > 1) and (S[Q-1] = ',') then Break;
          '"': if (Q > 1) and (S[Q-1] = '"') then Break
               else
               begin
                 Inc(Q, Max(Pos('"', Copy(S, Q+1, L-Q))+1, 1)); Numbers := False
               end;
          '(': begin
                 Inc(Q, Min(Pos(')', Copy(S, Q+1, L-Q))+1, 1)); Numbers := False
               end;
          else Inc(Q)
         end
       until Q > L;
      if Numbers and (Start <= L) and (Q > Start) then
       begin
         T := NumFormat.FindNumberStr(Copy(S, Start, Q - Start), V);
         if T = '' then
          begin
            Numbers := False;
            if AskAdd then
             begin
               T := Copy(S, Start, Q - Start);
               for N := 0 to AddStrNum - 1 do
                   if T = AddStr[N] then Numbers := True;
               if not Numbers and
                (MessageDLG('Allow "' + T + '" in tab ' + IntToStr(CollNum) +
                            ' of numbers line',
                            mtConfirmation, [mbYes,mbNo], 0) = 6) then
                begin
                  SetLength(AddStr, AddStrNum + 1); AddStr[AddStrNum] := T;
                  AddStrRow := AddStrRow + [Xm];
                  Inc(AddStrNum);
                  Numbers := True
                end
            end
          end
         else if Pos(T, S) <> Start then
          begin
            if (Pos(T, S) = 0) or (Pos(T, Copy(S, Start, Q - Start)) <> 1) then
               Numbers := False
          end
         else if (Start + Length(T) < Q)
         then
          if (NumFormat.FindNumber(Copy(S, Start+Length(T), Q-Start-Length(T)))
              <> 0)
          then Q := Start + Length(T)
          else if S[Q] <> ',' then Numbers := False
       end
    end
   else Q := Start + RecLen;
   Result := Copy(S, Start, Q - Start);
   while S[Q] = ' ' do Inc(Q);
   if S[Q] in [#9,','] then Start := Q + 1 else Start := Q;
   while S[Start] = ' ' do Inc(Start)
 end;
begin
  {$I-}
  AssignFile(F, FileName);
  FileMode := 0;  {Set file access to read only }
  Reset(F);
  {$I+}
  Result := IOResult = 0;
  if Result then
   begin
     ReadLn(F, S); L := Length(Trim(S));
     while (L = 0) and not EOF(F) do
           begin ReadLn(F, S); L := Length(Trim(S)) end;
     Result := L > 0;
     if not Result then CloseFile(F)
   end;
  if Result then
   begin
     if ANumFormat = Nil then NumFormat := TNumStringFormat.Create
                         else NumFormat := ANumFormat;
     L := Length(S);
     if Ydim > 0
     then begin DY := Table.FixedRows; Ym := Ydim + DY end
     else
      begin
        if -Ydim > Table.FixedRows
        then Ym := Abs(Ydim)
        else
         begin DY := GetFileSize; Ym := DY div L; if DY > L*Ym then Inc(Ym) end;
        DY := 0
      end;
     Numbers := True;
     Xm := Xdim;
     if Xm = 0 then
      begin
        if RecLen > 0 then
         begin
           Xm := L div RecLen; if L > RecLen*Xm then Inc(Xm)
         end
        else
         begin
           X := 1;
           repeat FindLabel(X, Xm, False); Inc(Xm) until X > L
         end;
        DX := 0;
        Table.ColCount := Xm;
      end
     else
      begin
        if Xm > 0
        then begin DX := Table.FixedCols; Table.ColCount := Xm + DX end
        else begin DX := 0; Xm := Abs(Xm); Table.ColCount := Xm end;
      end;
     Table.RowCount := Ym;
     Z := 1;
     if Numbers then
      begin
        if (DY = 1) and (Table.Cells[DX, 0] = '') then
           for X := 1 to Xm do Table.Cells[X + DX - 1, 0] := IntToStr(X);
        Z := 1; for X := 0 to Xm-1 do Table.Cells[X+DX, DY] := FindLabel(Z, X)
      end
     else with Table do
      begin
        if FixedRows = 0 then FixedRows := 1; DY := FixedRows;
        for Y := 0 to DY - 1 do
         begin
           Z := 1; for X := 0 to Xm-1 do Cells[X, Y] := FindLabel(Z, X);
           ReadLn(F, S)
         end;
        while not Numbers and not EOF(F) do
         begin
           Numbers := True; Z := 1; Xm := 0; X := Length(S); Y := DY;
           repeat
             Cells[Xm, Y] := FindLabel(Z, Xm); Inc(Xm)
           until Z > X;
           if not Numbers then begin DY := Y; FixedRows := DY; ReadLn(F, S) end
         end;
        if (Z < L) and (Ydim = 0) then
         begin RowCount := GetFileSize div Max(Z-1,1); Ym := RowCount - DY end
      end;
     SetLength(Values, Ym);
     if RecLen = 0 then RecLen := NumFormat.Width;
     for Y := 0 to Ym - 1 do with Table do
      begin
        SetLength(Values[Y], Xm);
        X := 0;
        while X < Xm do
         begin
           repeat
             if Trim(S) = '' then
             begin
               if not EOF(F) then
               repeat
                {$I-}
                ReadLn(F, S)
                {$I+}
               until (IOResult <> 0) or (Trim(S) <> '') or EOF(F);
               if Trim(S) = '' then Numbers := False
               else
               begin
                Z := 1;
                repeat until FindLabel(Z) <> ''
               end;
               if not Numbers then
               begin
                if (Xdim = 0) and (X > 0) then
                 begin
                   for Z := X+1 to Xm-1 do Values[Y,Z] := 0;
                   RowCount := Y + DY + 1;
                   SetLength(Values, Y + 1)
                 end
                else
                 begin
                   SetLength(Values[Y], 0);
                   RowCount := Y + DY;
                   SetLength(Values, Y)
                 end;
                CloseFile(F);
                if ANumFormat = Nil then NumFormat.Free;
                Exit
               end
             end;
             while S[1] = ' ' do Delete(S, 1, 1);
             if S[1] in [#9, ','] then
                begin Values[Y,X] := 0; T := ''; Break end
             else if (AddStrNum > 0) and (X in AddStrRow) then
              begin
                Z := 0; T := FindLabel(Z, X);
                if Numbers then Delete(S, 1, Length(T)) else Exit
              end
             else T := NumFormat.FindNumberStr(S, Values[Y,X], RecLen);
             if T = '' then S := '' else Break
           until False;
           Cells[X+DX, Y+DY] := T;
           Delete(S, 1, Pos(T, S) + Length(T) - 1);
           while (S <> '') and (S[1] = ' ') do Delete(S, 1, 1);
           if (S <> '') and (S[1] = ',') then Delete(S, 1, 1);
           if (S <> '') and (S[1] = #9) then Delete(S, 1, 1);
           Inc(X)
         end
      end;
     if Ydim = 0 then
      while not EOF(F) do
       begin
         ReadLn(F, S);
         if Trim(S) = '' then Break;
         Z := 1; FindLabel(Z); if not Numbers then Break;
         Table.RowCount := Ym + DY + 1;
         SetLength(Values, Ym + 1); SetLength(Values[Ym], Xm);
         for X := 0 to Xm - 1 do
          begin
            T := NumFormat.FindNumberStr(S, Values[Ym,X], RecLen);
            if T <> '' then
             begin
               Table.Cells[X+DX, Ym+DY] := T;
               Delete(S, 1, Pos(T, S) + Length(T) - 1)
             end
            else if Xdim <> 0 then
             begin
               SetLength(Values[Ym], 0); SetLength(Values, Ym);
               Table.RowCount := Ym + DY;
               Break
             end
          end;
         Inc(Ym)
       end;
     if ANumFormat = Nil then NumFormat.Free;
     CloseFile(F)
   end
end;

function FReadIntTab(const FileName: String;
         var Values: TInt2Array;
         var Table: TStringGrid;
         NumFormat: TNumStringFormat;
         Xdim, Ydim: Integer; RecLen: Integer = 0): Boolean;
var X, Y, Z: Integer; S: String; F: TextFile;
begin
  {$I-}
  AssignFile(F, FileName);
  FileMode := 0;  {Set file access to read only }
  Reset(F);
  {$I+}
  Result := IOResult = 0;
  if Result then
   begin
     Table.ColCount := Table.FixedCols + Xdim;
     Table.RowCount := Table.FixedRows + Ydim;
     SetLength(Values, Ydim);
     if RecLen = 0 then RecLen := NumFormat.Width;
     ReadLn(F, S);
     for Y := 0 to Ydim - 1 do with Table do
      begin
        SetLength(Values[Y], Xdim);
        X := 0;
        while X < Xdim do
         begin
           if Trim(S) = '' then
              if EOF(F) then
               begin
                 for Z := X to Xdim - 1 do
                  begin
                    Values[Y, Z] := Values[Y-1, Z];
                    Cells[Z+FixedCols, Y+FixedRows] :=
                      Cells[Z+FixedCols, Y+FixedRows-1]
                  end;
                 RowCount := Y + FixedRows + 1;
                 for Z := Y + 1 to Ydim - 1 do
                  begin
                    SetLength(Values[Z], Xdim);
                    for X := 0 to Xdim - 1 do Values[Z, X] := Values[Z-1, X]
                  end;
                 CloseFile(F);
                 Exit
               end
              else
               begin
                 {$I-}
                 ReadLn(F, S);
                 {$I+}
                 if IOResult <> 0 then
                  begin
                    for Z := X to Xdim - 1 do
                     begin
                       Values[Y, Z] := Values[Y-1, Z];
                       Cells[Z+FixedCols, Y+FixedRows] :=
                         Cells[Z+FixedCols, Y+FixedRows-1]
                     end;
                    RowCount := Y + FixedRows + 1;
                    for Z := Y + 1 to Ydim - 1 do
                     begin
                       SetLength(Values[Z], Xdim);
                       for X := 0 to Xdim - 1 do Values[Z, X] := Values[Z-1, X]
                     end;
                    CloseFile(F);
                    Exit
                  end
               end;
           Values[Y, X] := NumFormat.ExtractInteger(S, RecLen);
           Cells[X+FixedCols, Y+FixedRows] := IntToStr(Values[Y,X]);
           Inc(X)
         end
      end
   end;
  {$I-}
  CloseFile(F);
  {$I+}
  Z := IOResult
end;

function GetValue(const Data: array of Extended; X: Extended): Extended;
var P: Integer;
begin
  P := Floor(X);
  if P < 0
  then Result := EqLinExtrapol(Data[0], Data[1], X)
  else if P = 0
  then Result := GrEq3Interpol(Data[0], Data[1], Data[2], X)
  else if P < High(Data)-1
  then Result := Eq4Interpol(Data[P-1], Data[P], Data[P+1], Data[P+2], X-P)
  else if P = High(Data)-1
  then Result := GrEq3Interpol(Data[P+1], Data[P], Data[P-1], P+1-X)
  else
   begin
     P := High(Data); Result := EqLinExtrapol(Data[P], Data[P-1], P-X)
   end
end;

function ValueInCol(const Data: TReal2Array; Col: Integer; Y: Extended):
         Extended;
var P: Integer;
begin
  P := Floor(Y);
  if P < 0
  then Result := EqLinExtrapol(Data[0,Col], Data[1,Col], Y)
  else if P = 0
  then Result := GrEq3Interpol(Data[0,Col], Data[1,Col], Data[2,Col], Y)
  else if P < High(Data)-1
  then Result := Eq4Interpol(Data[P-1,Col], Data[P,Col],
                             Data[P+1,Col], Data[P+2,Col], Y-P)
  else if P = High(Data)-1
  then Result := GrEq3Interpol(Data[P+1,Col], Data[P,Col], Data[P-1,Col], P+1-Y)
  else
   begin
     P := High(Data); Result := EqLinExtrapol(Data[P,Col], Data[P-1,Col], P-Y)
   end
end;

function ValueInColPos(const Data: TReal2Array; const Ydat: array of Extended;
         Col: Integer; Y: Extended; Yid: Integer): Extended;
begin
  if Yid < 0
  then Result := LinExtrapol(Ydat[0],Data[0,Col], Ydat[1],Data[1,Col], Y)
  else if Yid = 0
  then Result := GrP3ClipInterp(Ydat[0],Data[0,Col], Ydat[1],Data[1,Col],
                              Ydat[2],Data[2,Col], Y)
  else if Yid < High(YDat)-1 then
   Result := P4ClipInterp(Ydat[Yid-1],Data[Yid-1,Col], Ydat[Yid],Data[Yid,Col],
                    Ydat[Yid+1],Data[Yid+1,Col], Ydat[Yid+2],Data[Yid+2,Col], Y)
  else if Yid = High(YDat)-1 then
   Result := GrP3ClipInterp(Ydat[Yid+1],Data[Yid+1,Col], Ydat[Yid],Data[Yid,Col],
                          Ydat[Yid-1],Data[Yid-1,Col], Y)
  else
   begin
    Yid := High(Data);
    Result := LinExtrapol(Ydat[Yid-1],Data[Yid-1,Col],Ydat[Yid],Data[Yid,Col],Y)
   end
end;

function ValueInCol(const Data: TReal2Array; const Ydat: array of Extended;
         Col: Integer; Y: Extended; var Yid: Integer): Extended;
begin
  GetPos(Ydat, Y, Yid);
  Result := ValueInColPos(Data, Ydat, Col, Y, Yid)
end;

function ValueInRow(const Data: TReal2Array; X: Extended; Row: Integer):
         Extended;
var P: Integer;
begin
  P := Floor(X);
  if P < 0
  then Result := EqLinExtrapol(Data[Row,0], Data[Row,1], X)
  else if P = 0
  then Result := GrEq3Interpol(Data[Row,0], Data[Row,1], Data[Row,2], X)
  else if P < High(Data)-1
  then Result := Eq4Interpol(Data[Row,P-1], Data[Row,P],
                             Data[Row,P+1], Data[Row,P+2], X-P)
  else if P = High(Data)-1
  then Result := GrEq3Interpol(Data[Row,P+1], Data[Row,P], Data[Row,P-1], P+1-X)
  else
   begin
     P := High(Data); Result := EqLinExtrapol(Data[Row,P], Data[Row,P-1], P-X)
   end
end;

function ValueInRowPos(const Data: TReal2Array; const Xdat: array of Extended;
         X: Extended; Row: Integer; Xid: Integer): Extended;
begin
  if Xid < 0
  then Result := LinExtrapol(Xdat[0],Data[Row,0], Xdat[1],Data[Row,1], X)
  else if Xid = 0
  then Result := GrP3ClipInterp(Xdat[0],Data[Row,0], Xdat[1],Data[Row,1],
                              Xdat[2],Data[Row,2], X)
  else if Xid < High(XDat)-1 then
   Result := P4ClipInterp(Xdat[Xid-1],Data[Row,Xid-1], Xdat[Xid],Data[Row,Xid],
                    Xdat[Xid+1],Data[Row,Xid+1], Xdat[Xid+2],Data[Row,Xid+2], X)
  else if Xid = High(XDat)-1 then
   Result := GrP3ClipInterp(Xdat[Xid+1],Data[Row,Xid+1], Xdat[Xid],Data[Row,Xid],
                          Xdat[Xid-1],Data[Row,Xid-1], X)
  else
   begin
    Xid := High(XDat);
    Result := LinExtrapol(Xdat[Xid-1],Data[Row,Xid-1],Xdat[Xid],Data[Row,Xid],X)
   end
end;

function ValueInRow(const Data: TReal2Array; const Xdat: array of Extended;
         X: Extended; Row: Integer; var Xid: Integer): Extended; overload;
begin
  GetPos(Xdat, X, Xid);
  Result := ValueInRowPos(Data, Xdat, X, Row, Xid)
end;

function GetValue(const Data: TReal2Array; X, Y: Extended): Extended;
var P: Integer;
begin
 if (Frac(X) = 0) and (X >= 0) and (X <= High(Data[0]))
 then Result := ValueInCol(Data, Round(X), Y)
 else
 begin
  P := Floor(Y);
  if (P=Y) and (P >= 0) and (P <= High(Data))
  then Result := ValueInRow(Data, X, P)
  else if P < 0
  then REsult := EqLinExtrapol(ValueInRow(Data, X,0), ValueInRow(Data, X,1), Y)
  else if P = 0
  then Result := GrEq3Interpol(ValueInRow(Data, X, 0), ValueInRow(Data, X, 1),
                               ValueInRow(Data, X, 2), Y)
  else if P < High(Data)-1
  then Result := Eq4Interpol(ValueInRow(Data, X, P-1), ValueInRow(Data, X, P),
                            ValueInRow(Data, X,P+1), ValueInRow(Data,X,P+2),Y-P)
  else if P = High(Data)-1
  then Result := GrEq3Interpol(ValueInRow(Data, X,P+1), ValueInRow(Data, X,P),
                               ValueInRow(Data, X,P-1), P+1-Y)
  else
   begin
    P := High(Data);
    Result := EqLinExtrapol(ValueInRow(Data, X,P), ValueInRow(Data, X,P-1), P-Y)
   end
 end
end;

function GetValue(const Data: TReal2Array; const Xdat, Ydat: array of Extended;
         X, Y: Extended; var Xid, Yid: Integer): Extended;
begin
 GetPos(Xdat, X, Xid); GetPos(Ydat, Y, Yid);
 if (Xid >= 0) and (Xdat[Xid] = X)
 then Result := ValueInColPos(Data, Ydat, Xid, Y, Yid)
 else
 begin
  if (Yid >= 0) and (Ydat[Yid] = Y)
  then Result := ValueInRowPos(Data, Xdat, X, Yid, Xid)
  else if Yid < 0
  then Result := LinExtrapol(Ydat[0], ValueInRowPos(Data, Xdat, X, 0, Xid),
                             Ydat[1], ValueInRowPos(Data, Xdat, X, 1, Xid), Y)
  else if Yid = 0
  then Result:= GrP3ClipInterp(Ydat[0], ValueInRowPos(Data, Xdat, X, 0, Xid),
                               Ydat[1], ValueInRowPos(Data, Xdat, X, 1, Xid),
                               Ydat[2], ValueInRowPos(Data, Xdat, X, 2, Xid), Y)
  else if Yid < High(Data)-1
  then Result:= P4ClipInterp(Ydat[Yid-1],ValueInRowPos(Data, Xdat, X, Yid-1,Xid),
                            Ydat[Yid], ValueInRowPos(Data, Xdat, X, Yid, Xid),
                            Ydat[Yid+1],ValueInRowPos(Data, Xdat, X, Yid+1,Xid),
                            Ydat[Yid+2],ValueInRowPos(Data, Xdat,X,Yid+2,Xid),Y)
  else if Yid = High(Data)-1
  then Result:=GrP3ClipInterp(Ydat[Yid+1],ValueInRowPos(Data, Xdat,X,Yid+1,Xid),
                              Ydat[Yid], ValueInRowPos(Data, Xdat, X, Yid, Xid),
                              Ydat[Yid-1],ValueInRowPos(Data,Xdat,X,Yid-1,Xid),Y)
  else
   begin
    Yid := High(Data);
    Result := LinExtrapol(Ydat[Yid-1], ValueInRowPos(Data, Xdat, X, Yid-1, Xid),
                          Ydat[Yid], ValueInRowPos(Data, Xdat, X, Yid, Xid), Y)
   end
 end
end;

procedure GetPos(const Data: array of Extended; Value: Extended;
         var Ipos: Integer);
// var Grouth: Boolean; Xa, Xi, X: Integer; Va, Vi, V: Extended;
begin
  if Data[High(Data)] > Data[0] then
  begin
    if Ipos = -1 then if Data[0] > Value then Exit else Ipos := 0;
    while (Data[Ipos] < Value) and (Ipos < High(Data)) do Inc(Ipos);
    while (Ipos >= 0) and (Data[Ipos] > Value) do Dec(Ipos)
  end
  else
  begin
    if Ipos = -1 then if Data[0] < Value then Exit else Ipos := 0;
    while (Data[Ipos] > Value) and (Ipos < High(Data)) do Inc(Ipos);
    while (Ipos >= 0) and (Data[Ipos] < Value) do Dec(Ipos)
  end;

 {Xa := Max(Min(Ipos, High(Data)-1), 0); Va := Data[Xa];
  Xi := Xa + 1; Vi := Data[Xi];
  if Va <= Vi then
    if Va <= Value then
       if Vi > Value then begin Ipos := Xa; Exit end
       else if (Vi=Value) or (Xi=High(Data)) then begin Ipos := Xi; Exit end
       else
        begin
          Inc(Xi); Vi := Data[Xi];
          if Vi >= Value then
             begin if Vi > Value then Dec(Xi); Ipos := Xi; Exit end
          else
           begin
             Xa := High(Data); Va := Data[Xa];
             if Va <= Value then begin Ipos := Xa; Exit end else Grouth := True
           end
        end
    else if Xa = 0 then begin Ipos := -1; Exit end
       else
        begin
          Dec(Xa); Va := Data[Xa];
          if Va <= Value then begin Ipos := Xa; Exit end
          else
           begin
             Xi := 0; Vi := Data[Xi];
             if Vi < Value then Grouth := True
             else begin if Vi = Value then Ipos := Xi else Ipos := -1; Exit end
           end
        end
  else if Va >= Value then
       if Vi < Value then begin Ipos := Xa; Exit end
       else if (Vi = Value) or (Xi = High(Data)) then begin Ipos := Xi; Exit end
       else
        begin
          Inc(Xi); Vi := Data[Xi];
          if Vi <= Value then
             begin if Vi < Value then Dec(Xi); Ipos := Xi; Exit end
          else
           begin
             Xa := High(Data); Va := Data[Xa];
             if Va >= Value then begin Ipos := Xa; Exit end else Grouth := False
           end
        end
    else if Xa = 0 then begin Ipos := -1; Exit end
       else
        begin
          Dec(Xa); Va := Data[Xa];
          if Va >= Value then begin Ipos := Xa; Exit end
          else
           begin
             Xi := 0; Vi := Data[Xi];
             if Vi > Value then Grouth := False
             else begin if Vi = Value then Ipos := Xi else Ipos := -1; Exit end
           end
        end;
  repeat
    X := Round(((Va*3.0 - Vi - Value*2.0)*Xi+
                (Va - Vi*3.0 + Value*2.0)*Xa)*0.25/(Va - Vi));
    V := Data[X];
    if V = Value then Break
    else if (V < Value) = Grouth then
     begin
       Xi := X+1; Vi := Data[Xi];
       if Vi = Value then begin Ipos := Xi; Exit end
       else if (Vi > Value) = Grouth then Break
     end
    else
     begin
       Xa := X-1; Va := Data[Xa];
       if (Va=Value) or ((Va < Value)=Grouth) then begin Ipos := Xa; Exit end
     end
  until False;
  Ipos := X}
end;

function FindValInCol(const Data: TReal2Array; Col: Integer; Value: Extended;
         InitRow: Integer = 0): Extended;
var Grouth: Boolean; Ya, Yi, Y: Integer; Va, Vi, V: Extended;
begin
  Y := InitRow; V := Data[Y,Col]; if V = Value then begin Result := Y; Exit end;
  if Y = 0 then Yi := Y else Yi := Y - 1; Vi := Data[Yi,Col];
  if Vi = Value then begin Result := Yi; Exit end;
  if Y = High(Data) then Ya := Y else Ya := Y + 1; Va := Data[Ya,Col];
  if Va = Value then begin Result := Ya; Exit end;
  if (Vi < Value) = (Va > Value) then Grouth := Va > Vi
  else
  begin
   Yi := 0; Vi := Data[Yi,Col]; if Vi = Value then begin Result := Yi; Exit end;
   Ya := High(Data); Va := Data[Ya,Col];
   if Va = Value then begin Result := Ya; Exit end;
   if Va >= Vi
   then if Value > Va then// Value > Va > Vi
           if Value < V then// V > Value > Va > Vi
              if Data[Y-1,Col] >= Data[Y+1,Col] then
                 begin Yi := Y; Vi := V; Grouth := False end
              else begin Ya := Y; Va := V; Grouth := True end
           else
           begin
            repeat
              V := Va; Y := Ya; Dec(Ya); Va := Data[Ya, Col]
            until Va < V;
            if Value >= Va then
             begin
               if Value > V
               then Result := LinExtrapol(V, Y, Va, Ya, Value)
               else if Y = High(Data)
               then Result := Y - InvGrEq3Interp(V, Va, Data[Ya-1, Col], Value)
               else
               Result :=
                 InvEq4Interp(Data[Ya-1,Col], Va, V, Data[Y+1,Col], Value) + Ya;
               Exit
             end
            else Grouth := True
           end
        else if Value < Vi then// Value < Vi < Va
           if Value > V then// V < Value < Vi < Va
              if Data[Y-1,Col] <= Data[Y+1,Col] then
                 begin Ya := Y; Va := V; Grouth := True end
              else begin Yi := Y; Vi := V; Grouth := False end
           else
           begin
            repeat
              V := Vi; Y := Yi; Inc(Yi); Vi := Data[Yi, Col]
            until Vi > V;
            if Value <= Vi then
             begin
               if Value <= V
               then Result := LinExtrapol(V, Y, Vi, Yi, Value)
               else if Y = 0
               then Result := InvGrEq3Interp(V, Vi, Data[Yi+1, Col], Value)
               else
               Result := InvEq4Interp(Data[Y-1,Col],V,Vi,Data[Yi+1,Col],Value)+Y;
               Exit
             end
            else Grouth := True
           end
        else Grouth := True
   else if Value <= Va then// Value < Va < Vi
           if Value > V then// V < Value < Va < Vi
              if Data[Y-1,Col] >= Data[Y+1,Col] then
                 begin Yi := Y; Vi := V; Grouth := False end
              else begin Ya := Y; Va := V; Grouth := True end
           else
           begin
            repeat
              V := Va; Y := Ya; Dec(Ya); Va := Data[Ya, Col]
            until Va > V;
            if Value <= Va then
             begin
               if Value <= V
               then Result := LinExtrapol(V, Y, Va, Ya, Value)
               else if Y = High(Data)
               then Result := Y - InvGrEq3Interp(V, Va, Data[Ya-1, Col], Value)
               else
               Result :=
                 InvEq4Interp(Data[Ya-1,Col], Va, V, Data[Y+1,Col], Value) + Ya;
               Exit
             end
            else Grouth := False
           end
        else if Value >= Vi then// Value > Vi > Va
           if Value < V then// V > Value > Vi > Va
              if Data[Y-1,Col] <= Data[Y+1,Col] then
                 begin Ya := Y; Va := V; Grouth := True end
              else begin Yi := Y; Vi := V; Grouth := False end
           else
           begin
            repeat
              V := Vi; Y := Yi; Inc(Yi); Vi := Data[Yi, Col]
            until Vi < V;
            if Value >= Vi then
             begin
               if Value >= V
               then Result := LinExtrapol(V, Y, Vi, Yi, Value)
               else if Y = 0
               then Result := InvGrEq3Interp(V, Vi, Data[Yi+1, Col], Value)
               else
               Result := InvEq4Interp(Data[Y-1,Col],V,Vi,Data[Yi+1,Col],Value)+Y;
               Exit
             end
            else Grouth := False
           end
        else Grouth := False
  end;
  if (InitRow > Yi) and (InitRow < Ya) then
   begin
     V := Data[InitRow, Col];
     if (V < Value) = Grouth
     then
      begin
        Yi := InitRow+1; Vi := Data[Yi, Col];
        if (Vi >= Value) = Grouth then
         begin
           if Yi = High(Data) then
            Result := Yi - InvGrEq3Interp(Vi, V, Data[InitRow-1,Col], Value)
           else
            Result := InvEq4Interp(Data[InitRow-1,Col],V,Vi,
                                   Data[Yi+1,Col],Value) + InitRow;
           Exit
         end
      end
     else
      begin
        Ya := InitRow-1; Va := Data[Ya, Col];
        if (Va <= Value) = Grouth then
         begin
           if Ya = 0 then
            Result := InvGrEq3Interp(Va, V, Data[InitRow+1,Col], Value)
          else
           Result := InvEq4Interp(Data[Ya-1,Col],Va,V,
                                  Data[InitRow+1,Col],Value) + Ya;
           Exit
         end
      end
   end;
  repeat
    Y := Round(((Va*3.0 - Vi - Value*2.0)*Yi+
                (Va - Vi*3.0 + Value*2.0)*Ya)*0.25/(Va - Vi));
    V := Data[Y, Col];
    if V = Value then Break
    else if (V < Value) = Grouth
    then
     begin
       Yi := Y+1; Vi := Data[Yi, Col];
       if Vi = Value then begin Result := Yi; Exit end
       else if (Vi > Value) = Grouth then
        begin
          if Y = 0 then
           Result := InvGrEq3Interp(V, Vi, Data[Yi+1,Col], Value)
          else if Yi = High(Data) then
           Result := Yi - InvGrEq3Interp(Vi, V, Data[Y-1,Col], Value)
          else
           Result := InvEq4Interp(Data[Y-1,Col],V,Vi,Data[Yi+1,Col],Value) + Y;
          Exit
        end
     end
    else
     begin
       Ya := Y-1; Va := Data[Ya, Col];
       if Va = Value then begin Result := Ya; Exit end
       else if (Va < Value) = Grouth then
        begin
          if Ya = 0 then
           Result := InvGrEq3Interp(Va, V, Data[Y+1,Col], Value)
          else if Y = High(Data) then
           Result := Y - InvGrEq3Interp(V, Va, Data[Ya-1,Col], Value)
          else
           Result := InvEq4Interp(Data[Ya-1,Col],Va,V,Data[Y+1,Col],Value) + Ya;
          Exit
        end
     end
  until False;
  Result := Y
end;

function FindValInCol(const Data: TReal2Array; const Ydat: array of Extended;
         Col: Integer; Value: Extended; var Row: Integer): Extended;
var Grouth: Boolean; Ya, Yi, Y, Y0, YH: Integer; V, Va, Vi, V0, VH, D: Extended;
begin
  YH := High(Data);
  Row := Max(Min(Row, YH), 0); V := Data[Row,Col];
  if V = Value then begin Yi := Row; Ya := Row end else
  begin
   Yi := Max(Row-1, 0); Vi := Data[Yi,Col];
   if Vi = Value then Ya := Yi else
   begin
    Ya := Min(Row+1, YH); Va := Data[Ya,Col];
    if Va = Value then Yi := Ya else
    begin
     VH := Data[YH,Col]; while Data[YH-1,Col] = VH do Dec(YH);
     if VH = Value then begin Yi := YH; Ya := YH end else
     begin
      Y0 := 0; V0 := Data[Y0,Col]; while Data[Y0+1,Col] = V0 do Inc(Y0);
      if V0 = Value then begin Yi := Y0; Ya := Y0 end else
      begin//          0 <= Y0 <= Yi=Row-1 - Row - Ya=Row+1 <= YH <= High(Data)
       Grouth := VH > V0;
       if ((Va > Vi) = Grouth) and (Va <> Vi) then//            V0(Vi) - (Va)VH
          if (Vi < Value) = (Value < Va) then//                 Vi - Value - Va
             if (Vi < Value) = (Value < V)
             then begin Va := V; Ya := Row end//Vi(Yi=Row-1) -Value- Va(Ya=Row)
             else begin Vi := V; Yi := Row end//Vi(Yi=Row) -Value- Va(Ya=Row+1)
          else if (Va < Value) = Grouth then//              (V0)Vi - Va - Value
           begin
             Y := (Ya + YH) div 2; V := Data[Y,Col];
             if V = Value then begin Yi := Y; Ya := Y end
             else if ((V < Va) = Grouth) or ((VH < Value) = Grouth) then
              begin//                                        V - Va(VH) - Value
                Vi := Va; Y := Ya;
                while (V <> Value) and ((V < Value) = Grouth) and
                      (V <> Vi) and ((Vi < V) = Grouth) and (Y < YH)
                   do begin Vi := V; Yi := Y; Inc(Y); V := Data[Y,Col] end;
                if V = Value then begin Yi := Y; Ya := Y end
                else if (Value < V) = Grouth then
                     begin Va := V; Ya := Y end//        Va(Y-1) - Value - V(Y)
                else begin Yi := YH; Ya := Yi + 1 end//           V(VH) - Value
              end
             else if (Value < V) = Grouth then//              Va - Value - V(Y)
                  begin Va := Vi; Yi := Ya; Ya := Y; Va := V end
             else begin Vi := V; Yi := Y; Ya := YH; Va := VH end//V(Y)-Value-VH
           end
          else//                                            Value - Vi - Va(VH)
           begin
             Y := (Yi + Y0) div 2; V := Data[Ya,Col];
             if V = Value then begin Yi := Y; Ya := Y end
             else if ((Vi < V) = Grouth) or ((V0 < Value) = Grouth) then
              begin//                                     Value - Vi(VH) - V(Y)
                Va := Vi; Ya := Yi;
                while (V <> Value) and ((Value < V) = Grouth) and
                      (V <> Va) and ((V < Va) = Grouth) and (Y > Y0)
                   do begin Va := V; Ya := Y; Dec(Y); V := Data[Y,Col] end;
                if V = Value then begin Yi := Y; Ya := Y end
                else if (V < Value) = Grouth then
                     begin Vi := V; Yi := Y end//        V(Y) - Value - Va(Y+1)
                else begin Yi := Y0-1; Ya := Y0 end//             Value - V(Vi)
              end
             else if (Value < V) = Grouth then//              V(Y) - Value - Vi
                  begin Ya := Yi; Yi := Y end
             else begin Vi := Y0; Ya := Y end//               V0 - Value - V(Y)
           end
       else if (VH < Value) = Grouth then//           (V0) Va - (Vi),VH - Value
             begin
                Y := Yi - 1; V := Data[Y,Col];
                while (V <> Vi) and ((Vi < V) = Grouth) and (Y0 < Y)
                   do begin Vi := V; Yi := Y; Dec(Y); V := Data[Y,Col] end;
                if V = Value then begin Yi := Y; Ya := Y end
                else if (V0 < Value) = (Value < Vi) then//    V0 - Value - V(Y)
                     begin Va := Vi; Ya := Yi; Yi := Y0 end
                else begin Yi := YH; Ya := YH + 1 end;//   (V0) V(Y),VH - Value
             end
       else  begin//                                                 Value - VH
                Y := Ya + 1; V := Data[Y,Col];
                while (V <> Va) and ((V < Va) = Grouth) and (Y < YH)
                   do begin Va := V; Ya := Y; Inc(Y); V := Data[Y,Col] end;
                if V = Value then begin Yi := Y; Ya := Y end
                else if (Va < Value) = (Value < VH) then//    V(Y) - Value - VH
                     begin Vi := Va; Yi := Ya; Ya := YH end
                else if (V0 < Value) = Grouth then
                     begin Yi := Y0-1; Ya := Y0 end//      Value - V0,V(Y) - VH
                else//                              V0 (Yi) - Value - V(Y) - VH
                 begin
                   Y := Yi - 1; V := Data[Y,Col];
                   while (V <> Vi) and ((Vi < V) = Grouth) and (Y0 < Y)
                   do begin Vi := V; Yi := Y; Dec(Y); V := Data[Y,Col] end;
                   if V = Value then begin Yi := Y; Ya := Y end
                   else begin Va := Vi; Ya := Yi; Yi := Y0; Vi := V0 end
                   //                                      V0 - Value - V(Y),VH
                 end
             end
      end
     end
    end
   end
  end;

  while Ya - Yi > 1 do
   begin
    Y := (Yi + Ya) div 2; V := Data[Y,Col];
    if V = Value then begin Yi := Y; Ya := Y end
    else if (V < Value) = Grouth then begin Vi := V; Yi := Y end//V(Y)-Value-Va
                                 else begin Va := V; Ya := Y end//Vi-Value-V(Y)
   end;

  Row := Yi;
  if Ya = Yi then Result := Ydat[Yi]
  else if Yi < Y0 then
       Result := LinExtrapol(V0,Ydat[Y0], Data[Y0+1,Col],Ydat[Y0+1], Value)
  else if Ya > YH then
       Result := LinExtrapol(Data[YH-1,Col],Ydat[YH-1], VH,Ydat[YH], Value)
  else if Yi = Y0 then Result := InvGrP3Interp(Ydat[Yi],Vi, Ydat[Ya],Va,
                                               Ydat[Ya+1],Data[Ya+1,Col], Value)
  else if Ya = YH then Result := InvGrP3Interp(Ydat[Ya],Va, Ydat[Yi],Vi,
                                               Ydat[Yi-1],Data[Yi-1,Col], Value)
  else Result := InvP4Interp(Ydat[Yi-1],Data[Yi-1,Col], Ydat[Yi],Vi,
                             Ydat[Ya],Va, Ydat[Ya+1],Data[Ya+1,Col], Value)
end;

function FindValInRow(const Data: TReal2Array; Row: Integer; Value: Extended;
         InitCol: Integer): Extended;
var Grouth: Boolean; Xa, Xi, X: Integer; Va, Vi, V: Extended;
begin
  Xi := 0; Vi := Data[Row,Xi]; if Vi = Value then begin Result := Xi; Exit end;
  Xa := High(Data); Va := Data[Row,Xa];
  if Va = Value then begin Result := Xa; Exit end;
  if Va >= Vi
  then if Value >= Va then
        begin
          repeat
            V := Va; X := Xa; Dec(Xa); Va := Data[Row, Xa]
          until Va < V;
          if Value >= Va then
           begin
             if Value >= V
             then Result := LinExtrapol(V, X, Va, Xa, Value)
             else if X = High(Data[Row])
             then Result := X - InvGrEq3Interp(V, Va, Data[Row, Xa-1], Value)
             else
             Result := InvEq4Interp(Data[Row,Xa-1],Va,V,Data[Row,X+1],Value)+Xa;
             Exit
           end
          else Grouth := True
        end
       else if Value <= Vi then
        begin
          repeat
            V := Vi; X := Xi; Inc(Xi); Vi := Data[Row, Xi]
          until Vi > V;
          if Value <= Vi then
           begin
             if Value <= V
             then Result := LinExtrapol(V, X, Vi, Xi, Value)
             else if X = 0
             then Result := InvGrEq3Interp(V, Vi, Data[Row, Xi+1], Value)
             else
              Result := InvEq4Interp(Data[Row,X-1],V,Vi,Data[Row,Xi+1],Value)+X;
             Exit
           end
          else Grouth := True
        end
       else Grouth := True
  else if Value <= Va then
        begin
          repeat
            V := Va; X := Xa; Dec(Xa); Va := Data[Row, Xa]
          until Va > V;
          if Value <= Va then
           begin
             if Value <= V
             then Result := LinExtrapol(V, X, Va, Xa, Value)
             else if X = High(Data[Row])
             then Result := X - InvGrEq3Interp(V, Va, Data[Row, Xa-1], Value)
             else
             Result := InvEq4Interp(Data[Row,Xa-1],Va,V,Data[Row,X+1],Value)+Xa;
             Exit
           end
          else Grouth := False
        end
       else if Value >= Vi then
        begin
          repeat
            V := Vi; X := Xi; Inc(Xi); Vi := Data[Row, Xi]
          until Vi < V;
          if Value >= Vi then
           begin
             if Value >= V
             then Result := LinExtrapol(V, X, Vi, Xi, Value)
             else if X = 0
             then Result := InvGrEq3Interp(V, Vi, Data[Row, Xi+1], Value)
             else
              Result := InvEq4Interp(Data[Row,X-1],V,Vi,Data[Row,Xi+1],Value)+X;
             Exit
           end
          else Grouth := False
        end
       else Grouth := False;
  if (InitCol > Xi) and (InitCol < Xa) then
   begin
     V := Data[Row, InitCol];
     if V = Value then begin Result := InitCol; Exit end
     else if (V < Value) = Grouth
     then
      begin
        Xi := InitCol+1; Vi := Data[Row, Xi];
        if (Vi >= Value) = Grouth then
         begin
           if Xi = High(Data) then
            Result := Xi - InvGrEq3Interp(Vi, V, Data[Row,InitCol-1], Value)
           else
            Result := InvEq4Interp(Data[Row,InitCol-1],V,Vi,
                                   Data[Row,Xi+1],Value) + InitCol;
           Exit
         end
      end
     else
      begin
        Xa := InitCol-1; Va := Data[Row, Xa];
        if (Va <= Value) = Grouth then
         begin
           if Xa = 0 then
            Result := InvGrEq3Interp(Va, V, Data[Row+1,InitCol+1], Value)
          else
           Result := InvEq4Interp(Data[Row, Xa-1],Va,V,
                                  Data[Row,InitCol+1],Value) + Xa;
           Exit
         end
      end
   end;
  repeat
    X := Round(((Va*3.0 - Vi - Value*2.0)*Xi+
                (Va - Vi*3.0 + Value*2.0)*Xa)*0.25/(Va - Vi));
    V := Data[Row, X];
    if V = Value then Break
    else if (V < Value) = Grouth
    then
     begin
       Xi := X+1; Vi := Data[Row, Xi];
       if Vi = Value then begin Result := Xi; Exit end
       else if (Vi > Value) = Grouth then
        begin
          if X = 0 then
           Result := InvGrEq3Interp(V, Vi, Data[Row,Xi+1], Value)
          else if Xi = High(Data[Row]) then
           Result := Xi - InvGrEq3Interp(Vi, V, Data[Row,X-1], Value)
          else
           Result := InvEq4Interp(Data[Row,X-1],V,Vi,Data[Row,Xi+1],Value) + X;
          Exit
        end
     end
    else
     begin
       Xa := X-1; Va := Data[Row, Xa];
       if Va = Value then begin Result := Xa; Exit end
       else if (Va <= Value) = Grouth then
        begin
          if Xa = 0 then
           Result := InvGrEq3Interp(Va, V, Data[Row,X+1], Value)
          else if X = High(Data[Row]) then
           Result := X - InvGrEq3Interp(V, Va, Data[Row,Xa-1], Value)
          else
           Result := InvEq4Interp(Data[Row,Xa-1],Va,V,Data[Row,X+1],Value) + Xa;
          Exit
        end
     end
  until False;
  Result := X
end;

function FindValInRow(const Data: TReal2Array; const Xdat: array of Extended;
         Row: Integer; Value: Extended; var Col: Integer): Extended;
var Grouth: Boolean; H, Xa, Xi, X: Integer; Va, Vi, V, V0: Extended;
begin
  H := High(Data[Row]);
  Xa := Max(Min(Col, H-1), 0); Va := Data[Row,Xa];
  if Va = Value then begin Col := Xa; Result := Xdat[Xa]; Exit end;
  Xi := Xa + 1; Vi := Data[Row,Xi];
  if Vi = Value then begin Col := Xi; Result := Xdat[Xi]; Exit end;
  V0 := Data[Row,0]; V := Data[Row,H]; Grouth := V > V0;
  if ((Va > Vi) = Grouth) or (Va = Vi) then
     if ((Value > V) = Grouth) or (Value = V) then
      begin
        if (Va = V) or ((Va > V) = Grouth) then begin Vi := V0; Xi := 0 end
                                           else begin Va := V; Xa := H end;
        repeat
          V := Va; X := Xa; Dec(Xa); Va := Data[Row,Xa]
        until (Va <> V) and ((Va < V) = Grouth);
        if Va = Value then Xi := Xa
        else if (Value > Va) = Grouth then
             if V = Value then begin Xi := X; Xa := X end
             else
              begin
                Xi := Xa; Xa := X; Vi := Va; Va := V;
                if (Value < Va) = Grouth then X := Xi
              end
      end
     else if ((Value < V0) = Grouth) or (Value = V0) then
      begin
        if ((Vi < V0) = Grouth) or (Vi = V0) then begin Va := V; Xa := H end
                                             else begin Vi := V0; Xi := 0 end;
        repeat
          V := Vi; X := Xi; Inc(Xi); Vi := Data[Row,Xi]
        until (V <> Vi) and ((V < Vi) = Grouth);
        if Vi = Value then Xa := Xi
        else if (Value < Vi) = Grouth then
             if V = Value then begin Xa := X; Xi := X end
             else
              begin
                Xa := Xi; Xi := X; Va := Vi; Vi := V;
                if (Value < Vi) = Grouth then Dec(X)
              end
      end
     else if (Value > Vi) = Grouth then begin Xa := H; Va := V end
     else begin Xi := 0; Vi := 0 end
  else if (Value > Vi) = Grouth then
     if ((Value > V) = Grouth) or (Value = V) then
        if (Xi < H) and ((Vi = V) or ((Vi > Value) = Grouth)) then
         begin
           X := Xi + 1; V := Data[Row,X];
           while (V <> Value) and ((V < Value) = Grouth) and
                 (V <> Vi) and ((Vi < V) = Grouth) do
            begin
              Va := Vi; Vi := V; Xa := Xi; Xi := X; Inc(X); V := Data[Row,X]
            end;
           if V = Value then begin Xi := X; Xa := X end
           else if ((V > Value) = Grouth) then begin Va:= V; Xa:= X; X:= Xi end
           else begin V := Vi; Vi := Va; Va := V; X := Xi; Xi := Xa; Xa := X end
         end
        else
         begin
           Xa := H; Va := V;
           repeat
             V := Va; X := Xa; Dec(Xa); Va := Data[Row,Xa]
           until (Va <> V) and ((Va < V) = Grouth);
           if Va = Value then Xi := Xa
           else if (Value > Va) = Grouth then
                if V = Value then begin Xi := X; Xa := X end
                else
                 begin
                   Xi := Xa; Xa := X; Vi := Va; Va := V;
                   if (Value < Va) = Grouth then X := Xi
                 end
         end
     else begin Xa := H; Va := V end
  else if (Value < Va) = Grouth then
     if ((Value < V0) = Grouth) or (Value = V0) then
        if (Xa > 0) and ((Va = V0) or ((Va < Value) = Grouth)) then
         begin
           X := Xa - 1; V := Data[Row,X];
           while (V <> Value) and ((V > Value) = Grouth) and
                 (V <> Va) and ((V < Va) = Grouth) do
            begin
              Vi := Va; Va := V; Xi := Xa; Xa := X; Dec(X); V := Data[Row,X]
            end;
           if V = Value then begin Xi := X; Xa := X end
           else if ((V < Value) = Grouth) then begin Vi := V; Xi := X end
           else begin V := Vi; Vi := Va; Va := V; H := Xi; Xi := Xa; Xa := H end
         end
        else
         begin
           Xi := 0; Vi := V0;
           repeat
             V := Vi; X := Xi; Inc(Xi); Vi := Data[Row,Xi]
           until (V <> Vi) and ((V < Vi) = Grouth);
           if Vi = Value then Xa := Xi
           else if (Value < Vi) = Grouth then
                if V = Value then begin Xi := X; Xa := X end
                else
                 begin
                   Xa := Xi; Xi := X; Va := Vi; Vi := V;
                   if (Value < Vi) = Grouth then Dec(X)
                 end
         end
     else begin Xi := 0; Vi := V0 end
  else begin V := Vi; Vi := Va; Va := V; X := Xa; Xa := Xi; Xi := X end;
  while Xa > Xi+1 do
   begin
    X := Round(((Va*3.0 - Vi - Value*2.0)*Xi+
                (Va - Vi*3.0 + Value*2.0)*Xa)*0.25/(Va - Vi));
    V := Data[Row,X];
    if V = Value then begin Xi := X; Xa := X end
    else if (V < Value) = Grouth
    then
     begin
       Xi := X+1; Vi := Data[Row,Xi];
       if Vi = Value then Xa := Xi
       else if (Vi > Value) = Grouth then
            begin Va := Vi; Vi := V; Xa := Xi; Xi := X end
     end
    else
     begin
       Xa := X-1; Va := Data[Row,Xa];
       if Va = Value then Xi := Xa
       else if (Va < Value) = Grouth then
            begin Vi := Va; Va := V; Xi := Xi; Xa := X; X := Xi end
     end
   end;
  if Xa = Xi then begin Col := Xa; Result := Xdat[Xa] end
  else
   begin
     Col := X;
     if X < Xi then Result := LinExtrapol(Vi,Xdat[Xi], Va,Xdat[Xa], Value)
     else if X = Xa then Result := LinExtrapol(Va,Xdat[Xa], Vi,Xdat[Xi], Value)
     else if Xi = 0 then Result := InvGrP3Interp(Xdat[Xi],Vi, Xdat[Xa],Va,
                                                 Xdat[3],Data[Row,3], Value)
     else if Xa = H then Result := InvGrP3Interp(Xdat[Xa],Va, Xdat[Xi],Vi,
                                              Xdat[Xi-1],Data[Row,Xi-1], Value)
     else Result := InvP4Interp(Xdat[Xi-1],Data[Row,Xi-1], Xdat[Xi],Vi,
                                Xdat[Xa],Va, Xdat[Xa+1],Data[Row,Xa+1], Value)
   end
end;

function FindValXpos(const Data: TReal2Array; Y, Value: Extended): Extended;
var P, X, Xa, Xi: Integer; Va, Vi, V: Extended; Grouth: Boolean;
begin
  if (Y >= 0) and (Y <= High(Data)) then
   begin
     P := Floor(Y); V := FindValInRow(Data, P, Value);
     if Frac(Y) = 0 then begin Result := V; Exit end;
     Vi := FindValInRow(Data, P+1, Value, Round(V));
     Xi := Floor(Min(V, Vi)); Xa := Min(Ceil(Max(V, Vi)), Xi + 1);
     Va := GetValue(Data, Xa, Y); Vi := GetValue(Data, Xi, Y);
     Grouth := Va > Vi
   end
  else
   begin
     if Y < 0 then
      begin
        P := 0; V := FindValInRow(Data, 0, Value);
        Vi := LinExtrapol(0, V, 1, FindValInRow(Data, 1, Value, Round(V)), Y)
      end
     else
      begin
        P := High(Data); V := FindValInRow(Data, P, Value);
        Vi := LinExtrapol(0,V, -1,FindValInRow(Data, P-1, Value, Round(V)), Y-P)
      end;
     Xi := Floor(Min(V, Vi)); Xa := Min(Ceil(Max(V, Vi)), Xi+1);
     Va := GetValue(Data, Xa, Y); Vi := GetValue(Data, Xi, Y);
     while (Vi = Va) and (Xi > 0) do
           begin Dec(Xi); Vi := GetValue(Data, Xi, Y) end;
     while (Vi = Va) and (Xa < High(Data[P])) do
           begin Inc(Xa); Va := GetValue(Data, Xa, Y) end;
     if Vi < Va then
      begin
        while Vi > Value do
         begin
           V := Vi; Dec(Xi); Vi := GetValue(Data, Xi, Y); if Vi >= V then Break
         end;
        if Vi = Value then begin Result := Xi; Exit end;
        while Va < Value do
         begin
           V := Va; Inc(Xa); Va := GetValue(Data, Xa, Y); if Va <= V then Break
         end;
        if Va = Value then begin Result := Xa; Exit end;
        Grouth := True;
      end
     else
      begin
        while Vi < Value do
         begin
           V := Vi; Dec(Xi); Vi := GetValue(Data, Xi, Y); if Vi <= V then Break
         end;
        if Vi = Value then begin Result := Xi; Exit end;
        while Va > Value do
         begin
           V := Va; Inc(Xa); Va := GetValue(Data, Xa, Y); if Va >= V then Break
         end;
        if Va = Value then begin Result := Xa; Exit end;
        Grouth := False
      end
   end;
  if Xa <= 0
  then
   begin
     Result := LinExtrapol(ValueInCol(Data, 1, Y), 1,
                           ValueInCol(Data, 0, Y), 0, Value);
     Exit
   end
  else if Xi >= High(Data[P])
  then
   begin
     P := High(Data[P]);
     Result := LinExtrapol(ValueInCol(Data, P-1, Y), -1,
                           ValueInCol(Data, P, Y), 0, Value) + P;
     Exit
   end
  else if Xa - Xi = 1
  then
   begin
     if Xi = 0
     then Result := InvGrEq3Interp(Vi, Va, ValueInCol(Data,2,Y), Value)
     else if Xa = High(Data[P])
     then Result := Xa - InvGrEq3Interp(Va, Vi, ValueInCol(Data,Xa-2,Y), Value)
     else Result := InvEq4Interp(ValueInCol(Data, Xi-1, Y), Vi, Va,
                                 ValueInCol(Data, Xa+1, Y), Value) + Xi;
     Exit
   end;
  repeat
    X := Round(((Va*3.0 - Vi - Value*2.0)*Xi+
                (Va - Vi*3.0 + Value*2.0)*Xa)*0.25/(Va - Vi));
    V := GetValue(Data, X, Y);
    if V = Value then Break
    else if (V < Value) = Grouth
    then
     begin
       Xi := X+1; Vi := GetValue(Data, Xi, Y);
       if Vi = Value then begin Result := Xi; Exit end
       else if (Vi > Value) = Grouth then
        begin
          if X = 0
          then Result := InvGrEq3Interp(V, Vi, ValueInCol(Data, Xi+1, Y), Value)
          else if Xi = High(Data[P])
          then Result := Xi - InvGrEq3Interp(Vi,V,ValueInCol(Data,X-1,Y), Value)
          else Result := InvEq4Interp(ValueInCol(Data, X-1, Y), V, Vi,
                                      ValueInCol(Data, Xi+1, Y), Value) + X;
          Exit
        end
     end
    else
     begin
       Xa := X-1; Va := GetValue(Data, Xa, Y);
       if Va = Value then begin Result := Xa; Exit end
       else if (Va <= Value) = Grouth then
        begin
          if Xa = 0
          then Result := InvGrEq3Interp(Va, V, ValueInCol(Data, X+1, Y), Value)
          else if X = High(Data[P])
          then Result := X - InvGrEq3Interp(V,Va,ValueInCol(Data,Xa-1,Y), Value)
          else Result := InvEq4Interp(ValueInCol(Data, Xa-1, Y), Va, V,
                                      ValueInCol(Data, X+1, Y), Value) + Xa;
          Exit
        end
     end
  until Xa > Xi;
  Result := X
end;

function FindValXpos(const Data: TReal2Array;
         const Xdat, Ydat: array of Extended;
         Y, Value: Extended; var Xtab, Ytab: Integer): Extended;
var R, H, X, Xa, Xi: Integer; V, Va, Vi, Vq: Extended; Grouth : Boolean;
begin
  GetPos(Ydat, Y, Ytab); R := Max(Ytab, 0); Xtab := Max(Min(Xtab,High(Xdat)),0);
  Vi := FindValInRow(Data, Xdat, R, Value, Xtab);
  if Y = Ydat[Ytab] then begin Result := Vi; Exit end;
  Xa := Xtab;
  H := High(Data);
  if R = H then
     begin Va := Vi; Dec(R); Vi := FindValInRow(Data, Xdat, R, Value, Xtab) end
  else Va := FindValInRow(Data, Xdat, R+1, Value, Xa);
  if (Ytab < 0) or (Ytab = H) then
   begin
     {X := Round(} V := LinExtrapol(Ydat[R],Xtab, Ydat[R+1],Xa, Y){)};
     GetPos(Xdat, LinExtrapol(Ydat[R],{Vi}Xtab, Ydat[R+1],{Va}Xa, V), X);
     Xi := Max(Min(Min(Xa, Xtab), X), 0);
     Xa := Min(Max(Max(Xa, Xtab), X), High(Xdat))
   end
  else if Xa >= Xtab then Xi := Xtab else begin Xi := Xa; Xa := Xtab end;

  H := High(Data[R]);
  if Xi = Xa then
     if Xa = H then
      begin
        Va := ValueInColPos(Data, Ydat, Xa, Y, Ytab);
        if Va <> Value then
         begin
           Dec(Xi); Vi := ValueInColPos(Data, Ydat, Xi, Y, Ytab);
           if Vi = Value then Xa := Xi
           else if Vi = Va then Xi := Xa
           else if (Vi < Value) = (Va > Value) then X := Xi else X := Xa
         end
      end
     else
      begin
        if Xi < 0 then Xi := 0;
        Vi := ValueInColPos(Data, Ydat, Xi, Y, Ytab);
        if Vi = Value then Xa := Xi
        else
         begin
           Xa := Xi+1; Va := ValueInColPos(Data, Ydat, Xa, Y, Ytab);
         end;
        if Va = Value then Xi := Xa
        else if Va = Vi then Xa := Xi
        else if (Vi < Value) = (Va > Value) then X := Xi else X := Xi - 1
      end
  else
   begin
     if Xi < 0 then begin Xi := 0; if Xa < 1 then Xa := 1 end;
     Vi := ValueInColPos(Data, Ydat, Xi, Y, Ytab);
     if Vi = Value then Xa := Xi else
      begin
        Va := ValueInColPos(Data, Ydat, Xa, Y, Ytab);
        if Va = Value then Xi := Xa else
         begin
           Grouth := Data[R,0] < Data[R,H];
           if (Va < Value) = Grouth then
            begin
              if Xa < H then
               begin
                 Vq := ValueInColPos(Data, Ydat, H, Y, Ytab);
                 X := Xa + 1; V := ValueInColPos(Data, Ydat, X, Y, Ytab);
                 while (((V <> Va) and ((Va < V) = Grouth)) or
                        ((V <> Vq) and ((V < Vq) = Grouth))) and
                       (V <> Value) and ((V < Value) = Grouth) and (X < H) do
                  begin
                    Vi := Va; Va := V; Xi := Xa; Xa := X;
                    Inc(X); V := ValueInColPos(Data, Ydat, X, Y, Ytab)
                  end;
                 if V = Value then begin Xtab := X; Result := Xdat[X]; Exit end
                 else if (V > Value) = Grouth then
                  begin
                    Xtab := Xa;
                    if X < H then
                     Result := InvP4Interp(Xdat[Xi],Vi, Xdat[Xa],Va, Xdat[X],V,
                       Xdat[X+1],ValueInColPos(Data, Ydat, X+1, Y, Ytab), Value)
                    else Result := InvGrP3Interp(Xdat[X],V, Xdat[Xa],Va,
                                                 Xdat[Xi],Vi, Value);
                    Exit
                  end
                 else if (V <> Va) and ((Va < V) = Grouth) then
                  begin
                    Xtab := X;
                    Result := LinExtrapol(V,Xdat[X], Va,Xdat[Xa], Value);
                    Exit
                  end
               end;
              if (Vi < Value) = Grouth then
               begin
                 if Xa > Xi + 1 then
                  begin
                    Xa := Xi + 1; Va := ValueInColPos(Data, Ydat, Xa, Y, Ytab)
                  end;
                 if (Vi = Va) or ((Vi > Va) = Grouth) then
                  repeat
                    Va := Vi; Xa := Xi;
                    Dec(Xi); Vi := ValueInColPos(Data, Ydat, Xi, Y, Ytab)
                  until (Vi <> Va) and ((Vi < Va) = Grouth)
                 else if (Xa < H) and (Va <> Value) and ((Va < Value) = Grouth)
                 then
                  begin
                    X := Xa + 1; V := ValueInColPos(Data, Ydat, X, Y, Ytab);
                    while (V <> Va) and ((Va < V) = Grouth) do
                     begin
                       Vi := Va; Va := V; Xi := Xa; Xa := X;
                       Inc(X); V := ValueInColPos(Data, Ydat, X, Y, Ytab);
                       if (V = Value) or ((Value < V) = Grouth) then
                        begin Vi := Va; Va := V; Xi := Xa; Xa := X; Break end
                     end
                  end;
                 if Vi = Value then Xa := Xi else if Va = Value then Xi := Xa
                 else if ((Va < Value) = Grouth) then X := Xa else X := Xi
               end
            end
           else X := Xi;
           if (Vi > Value) = Grouth then
            begin
              if Xi > 0 then
               begin
                 Vq := ValueInColPos(Data, Ydat, 0, Y, Ytab);
                 X := Xi - 1; V := ValueInColPos(Data, Ydat, X, Y, Ytab);
                 while (((V <> Vi) and ((V < Vi) = Grouth)) or
                        ((V <> Vq) and ((Vq < V) = Grouth))) and
                       (V <> Value) and ((V > Value) = Grouth) and (X > 0) do
                  begin
                    Va := Vi; Vi := V; Xa := Xi; Xi := X;
                    Dec(X); V := ValueInColPos(Data, Ydat, X, Y, Ytab)
                  end;
                 if V = Value then begin Xtab := X; Result := Xdat[X]; Exit end
                 else if (V < Value) = Grouth then
                  begin
                    Xtab := X;
                    if X > 0 then
                     Result := InvP4Interp(Xdat[X-1],
                                    ValueInColPos(Data, Ydat, X-1, Y, Ytab),
                                    Xdat[X],V, Xdat[Xi],Vi, Xdat[Xa],Va, Value)
                    else Result := InvGrP3Interp(Xdat[X],V, Xdat[Xi],Vi,
                                                 Xdat[Xa],Va, Value);
                    Exit
                  end
                 else if (V <> Vi) and ((V < Vi) = Grouth) then
                  begin
                    Xtab := X - 1;
                    Result := LinExtrapol(V,Xdat[X], Vi,Xdat[Xi], Value);
                    Exit
                  end
               end;
              if Xi < Xa - 1 then
               begin
                 Xi := Xa - 1; Vi := ValueInColPos(Data, Ydat, Xi, Y, Ytab)
               end;
              if (Vi = Va) or ((Vi > Va) = Grouth) then
               repeat
                 Vi := Va; Xi := Xa;
                 Inc(Xa); Va := ValueInColPos(Data, Ydat, Xa, Y, Ytab)
               until (Vi <> Va) and ((Vi < Va) = Grouth)
              else if (Xi > 0) and (Vi <> Value) and ((Vi > Value) = Grouth)
                 then
                  begin
                    X := Xi - 1; V := ValueInColPos(Data, Ydat, X, Y, Ytab);
                    while (V <> Vi) and ((V < Vi) = Grouth) do
                     begin
                       Va := Vi; Vi := V; Xa := Xi; Xi := X;
                       Dec(X); V := ValueInColPos(Data, Ydat, X, Y, Ytab);
                       if (V = Value) or ((Value > V) = Grouth) then
                        begin Va := Vi; Vi := V; Xa := Xi; Xi := X; Break end
                     end
                  end;
              if Vi = Value then Xa := Xi else if Va = Value then Xi := Xa
              else if ((Vi > Value) = Grouth) then X := Xi-1 else X := Xi
            end
         end
      end
   end;

  while Xa > Xi + 1 do
   begin
    X := Round(((Va*3.0 - Vi - Value*2.0)*Xi+
                (Va - Vi*3.0 + Value*2.0)*Xa)*0.25/(Va - Vi));
    V := ValueInColPos(Data, Ydat, X, Y, Ytab);
    if V = Value then begin Xi := X; Xa := X end
    else if (V < Value) = Grouth then
     begin
       Xi := X+1; Vi := ValueInColPos(Data, Ydat, Xi, Y, Ytab);
       if Vi = Value then Xa := Xi
       else if (Vi > Value) = Grouth then
            begin Xa := Xi; Xi := X; Va := Vi; Vi := V end
     end
    else
     begin
       Xa := X-1; Va := ValueInColPos(Data, Ydat, Xa, Y, Ytab);
       if Va = Value then Xi := Xa
       else if (Va < Value) = Grouth then
            begin Xi := Xa; Xa := X; X := Xi; Vi := Va; Va := V end
     end
   end;

   if Xi = Xa then begin Xtab := Xi; Result := Xdat[Xi] end
   else
    begin
      Xtab := X;
      if X < Xi then Result := LinExtrapol(Vi,Xdat[Xi], Va,Xdat[Xa], Value)
      else if X >= Xa then Result := LinExtrapol(Va,Xdat[Xa],Vi,Xdat[Xi], Value)
      else if Xi = 0 then
           Result := InvGrP3Interp(Xdat[Xi],Vi, Xdat[Xa],Va,
                          Xdat[3],ValueInColPos(Data, Ydat, 3, Y, Ytab), Value)
      else if Xa = H then
           Result := InvGrP3Interp(Xdat[Xa],Va, Xdat[Xi], Vi,
                          Xdat[H-2],ValueInColPos(Data,Ydat,H-2,Y,Ytab), Value)
      else Result := InvP4Interp(Xdat[Xi-1],ValueInColPos(Data,Ydat,Xi-1,Y,Ytab),
                                 Xdat[Xi],Vi, Xdat[Xa], Va,
                                 Xdat[Xa+1], ValueInColPos(Data,Ydat,Xa+1,Y,Ytab),
                                 Value)
    end
end;

function FindValYpos(const Data: TReal2Array; X, Value: Extended): Extended;
var P, Y, Ya, Yi: Integer; Va, Vi, V: Extended; Grouth: Boolean;
begin
  if (X >= 0) and (X <= High(Data[0])) then
   begin
     P := Floor(X); V := FindValInCol(Data, P, Value);
     if Frac(X) = 0 then begin Result := V; Exit end;
     Vi := FindValInCol(Data, P+1, Value, Round(V));
     Yi := Floor(Min(V, Vi)); Ya := Min(Ceil(Max(V, Vi)), Yi + 1);
     Va := GetValue(Data, X, Ya); Vi := GetValue(Data, X, Yi);
     Grouth := Va > Vi
   end
  else
   begin
     if X < 0 then
      begin
        P := 0; V := FindValInCol(Data, 0, Value);
        Vi := LinExtrapol(0, V, 1, FindValInCol(Data, 1, Value, Round(V)), X)
      end
     else
      begin
        P := High(Data[0]); V := FindValInCol(Data, P, Value);
        Vi := LinExtrapol(0,V, -1,FindValInCol(Data, P-1, Value, Round(V)), X-P)
      end;
     Yi := Floor(Min(V, Vi)); Ya := Min(Ceil(Max(V, Vi)), Yi+1);
     Va := GetValue(Data, X, Ya); Vi := GetValue(Data, X, Yi);
     while (Vi = Va) and (Yi > 0) do
           begin Dec(Yi); Vi := GetValue(Data, X, Yi) end;
     while (Vi = Va) and (Ya < High(Data)) do
           begin Inc(Ya); Va := GetValue(Data, X, Ya) end;
     if Vi < Va then
      begin
        while Vi > Value do
         begin
           V := Vi; Dec(Yi); Vi := GetValue(Data, X, Yi); if Vi >= V then Break
         end;
        if Vi = Value then begin Result := Yi; Exit end;
        while Va < Value do
         begin
           V := Va; Inc(Ya); Va := GetValue(Data, X, Ya); if Va <= V then Break
         end;
        if Va = Value then begin Result := Ya; Exit end;
        Grouth := True;
      end
     else
      begin
        while Vi < Value do
         begin
           V := Vi; Dec(Yi); Vi := GetValue(Data, X, Yi); if Vi <= V then Break
         end;
        if Vi = Value then begin Result := Yi; Exit end;
        while Va > Value do
         begin
           V := Va; Inc(Ya); Va := GetValue(Data, X, Ya); if Va >= V then Break
         end;
        if Va = Value then begin Result := Ya; Exit end;
        Grouth := False
      end
   end;
  if Ya <= 0
  then
   begin
     Result := LinExtrapol(ValueInRow(Data, X, 1), 1,
                           ValueInRow(Data, X, 0), 0, Value);
     Exit
   end
  else if Yi >= High(Data)
  then
   begin
     P := High(Data);
     Result := LinExtrapol(ValueInRow(Data, X, P-1), -1,
                           ValueInRow(Data, X, P), 0, Value) + P;
     Exit
   end
  else if Ya - Yi = 1
  then
   begin
     if Yi = 0
     then Result := InvGrEq3Interp(Vi, Va, ValueInRow(Data,X,2), Value)
     else if Ya = High(Data)
     then Result := Ya - InvGrEq3Interp(Va, Vi, ValueInRow(Data,X,Ya-2), Value)
     else Result := InvEq4Interp(ValueInRow(Data, X, Yi-1), Vi, Va,
                                 ValueInRow(Data, X, Ya+1), Value) + Yi;
     Exit
   end;
  repeat
    Y := Round(((Va*3.0 - Vi - Value*2.0)*Yi+
                (Va - Vi*3.0 + Value*2.0)*Ya)*0.25/(Va - Vi));
    V := GetValue(Data, X, Y);
    if V = Value then Break
    else if (V < Value) = Grouth
    then
     begin
       Yi := Y+1; Vi := GetValue(Data, X, Yi);
       if Vi = Value then begin Result := Yi; Exit end
       else if (Vi > Value) = Grouth then
        begin
          if Y = 0
          then Result := InvGrEq3Interp(V, Vi, ValueInRow(Data, X, Yi+1), Value)
          else if Yi = High(Data)
          then Result := Yi - InvGrEq3Interp(Vi,V,ValueInRow(Data,X,Y-1), Value)
          else Result := InvEq4Interp(ValueInRow(Data, X, Y-1), V, Vi,
                                      ValueInRow(Data, X, Yi+1), Value) + Y;
          Exit
        end
     end
    else
     begin
       Ya := Y-1; Va := GetValue(Data, X, Ya);
       if Va = Value then begin Result := Ya; Exit end
       else if (Va <= Value) = Grouth then
        begin
          if Ya = 0
          then Result := InvGrEq3Interp(Va, V, ValueInRow(Data, X, Y+1), Value)
          else if Y = High(Data)
          then Result := Y - InvGrEq3Interp(V,Va,ValueInRow(Data,X,Ya-1), Value)
          else Result := InvEq4Interp(ValueInRow(Data, X, Ya-1), Va, V,
                                      ValueInRow(Data, X, Y+1), Value) + Ya;
          Exit
        end
     end
  until Ya > Yi;
  Result := Y
end;

function FindValYpos(const Data: TReal2Array;
         const Xdat, Ydat: array of Extended;
         X, Value: Extended; var Xtab, Ytab: Integer): Extended;
         overload;
var R, H, Y, Ya, Yi: Integer; Va, Vi, V, Vq: Extended; Grouth: Boolean;
begin
  GetPos(Xdat, X, Xtab); R := Max(Xtab, 0);
  // Ytab := Max(Min(Ytab,High(Ydat)),0);
  Vi := FindValInCol(Data, Ydat, R, Value, Ytab);
  if X = Xdat[Xtab] then begin Result := Vi; Exit end;
  Ya := Ytab;
  H := High(Xdat);
  if R = H then
     begin Va := Vi; Dec(R); Vi := FindValInCol(Data, Ydat, R, Value, Ytab) end
  else Va := FindValInCol(Data, Ydat, R+1, Value, Ya);
  if (Xtab < 0) or (Xtab = H) then
   begin
     {Y := Round(} V := LinExtrapol(Xdat[R],Ytab, Xdat[R+1],Ya, X){)};
     GetPos(Ydat, LinExtrapol(Xdat[R],{Vi}Ytab, Xdat[R+1],{Va}Ya, V), Y);
     Yi := Max(Min(Min(Ya, Ytab), Y), 0);
     Ya := Min(Max(Max(Ya, Ytab), Y), High(YDat))
   end
  else if Ya >= Ytab then Yi := Ytab else begin Yi := Ya; Ya := Ytab end;

  H := High(YDat);
  if Yi = Ya then
     if Ya = H then
      begin
        Va := ValueInRowPos(Data, Xdat, X, Ya, Xtab);
        if Va <> Value then
         begin
           Dec(Yi); Vi := ValueInRowPos(Data, Xdat, X, Yi, Xtab);
           if Vi = Value then Ya := Yi
           else if Vi = Va then Yi := Ya
           else if (Vi < Value) = (Va > Value) then Y := Yi else Y := Ya
         end
      end
     else
      begin
        if Yi < 0 then Yi := 0;
        Vi := ValueInRowPos(Data, Xdat, X, Yi, Xtab);
        if Vi = Value then Ya := Yi
        else
         begin
           Ya := Yi+1; Va := ValueInRowPos(Data, Xdat, X, Ya, Xtab);
         end;
        if Va = Value then Yi := Ya
        else if Va = Vi then Ya := Yi
        else if (Vi < Value) = (Va > Value) then Y := Yi else Y := Yi - 1
      end
  else
   begin
     if Yi < 0 then begin Yi := 0; if Ya < 1 then Ya := 1 end;
     Vi := ValueInRowPos(Data, Xdat, X, Yi, Xtab);
     if Vi = Value then Ya := Yi else
      begin
        Va := ValueInRowPos(Data, Xdat, X, Ya, Xtab);
        if Va = Value then Yi := Ya else
         begin
           Grouth := Data[0,R] < Data[H,R];
           if (Va < Value) = Grouth then
            begin
              if Ya < H then
               begin
                 Vq := ValueInRowPos(Data, Xdat, X, H, Xtab);
                 Y := Ya + 1; V := ValueInRowPos(Data, Xdat, X, Y, Xtab);
                 while (((V <> Va) and ((Va < V) = Grouth)) or
                        ((V <> Vq) and ((V < Vq) = Grouth))) and
                       (V <> Value) and ((V < Value) = Grouth) and (Y < H) do
                  begin
                    Vi := Va; Va := V; Yi := Ya; Ya := Y;
                    Inc(Y); V := ValueInRowPos(Data, Xdat, X, Y, Xtab)
                  end;
                 if V = Value then begin Ytab := Y; Result := Ydat[Y]  end
                 else if (V > Value) = Grouth then
                  begin
                    Ytab := Ya;
                    if Y < H then
                     Result := InvP4Interp(Ydat[Yi],Vi, Ydat[Ya],Va, Ydat[Y],V,
                       Ydat[Y+1],ValueInRowPos(Data, Xdat, X, Y+1, Xtab), Value)
                    else Result := InvGrP3Interp(Ydat[Y],V, Ydat[Ya],Va,
                                                 Ydat[Yi],Vi, Value);
                    Exit
                  end
                 else if (V <> Va) and ((Va < V) = Grouth) then
                  begin
                    Ytab := Y;
                    Result := LinExtrapol(V,Ydat[Y], Va,Ydat[Ya], Value);
                    Exit
                  end
               end;
              if (Vi < Value) = Grouth then
               begin
                 if Ya > Yi + 1 then
                  begin
                    Ya := Yi + 1; Va := ValueInRowPos(Data, Xdat, X, Ya, Xtab)
                  end;
                 if (Vi = Va) or ((Vi > Va) = Grouth) then
                  repeat
                    Va := Vi; Ya := Yi;
                    Dec(Yi); Vi := ValueInRowPos(Data, Xdat, X, Yi, Xtab)
                  until (Vi <> Va) and ((Vi < Va) = Grouth)
                 else if (Ya < H) and (Va <> Value) and ((Va < Value) = Grouth)
                 then
                  begin
                    Y := Ya + 1; V := ValueInRowPos(Data, Xdat, X, Y, Xtab);
                    while (V <> Va) and ((Va < V) = Grouth) do
                     begin
                       Vi := Va; Va := V; Yi := Ya; Ya := Y;
                       Inc(Y); if Y > H then Break;
                       V := ValueInRowPos(Data, Xdat, X, Y, Xtab);
                       if (V = Value) or ((Value < V) = Grouth) then
                        begin Vi := Va; Va := V; Yi := Ya; Ya := Y; Break end
                     end
                  end;
                 if Vi = Value then Ya := Yi else if Va = Value then Yi := Ya
                 else if ((Va < Value) = Grouth) then Y := Ya else Y := Yi
               end
            end
           else Y := Yi;
           if (Vi > Value) = Grouth then
            begin
              if Yi > 0 then
               begin
                 Vq := ValueInRowPos(Data, Xdat, X, 0, Xtab);
                 Y := Yi - 1; V := ValueInRowPos(Data, Xdat, X, Y, Xtab);
                 while (((V <> Vi) and ((V < Vi) = Grouth)) or
                        ((V <> Vq) and ((Vq < V) = Grouth))) and
                       (V <> Value) and ((V > Value) = Grouth) and (Y > 0) do
                  begin
                    Va := Vi; Vi := V; Ya := Yi; Yi := Y;
                    Dec(Y); V := ValueInRowPos(Data, Xdat, X, Y, Xtab);
                  end;
                 if V = Value then begin Ytab := Y; Result := Ydat[Y]  end
                 else if (V < Value) = Grouth then
                  begin
                    Ytab := Y;
                    if Y > 0 then
                     Result := InvP4Interp(Ydat[Y-1],
                                    ValueInRowPos(Data, Xdat, X, Y-1, Xtab),
                                    Ydat[Y],V, Ydat[Yi],Vi, Ydat[Ya],Va, Value)
                    else Result := InvGrP3Interp(Ydat[Y],V, Ydat[Yi],Vi,
                                                 Ydat[Ya],Va, Value);
                    Exit
                  end
                 else if (V <> Vi) and ((V < Vi) = Grouth) then
                  begin
                    Ytab := Y - 1;
                    Result := LinExtrapol(V,Ydat[Y], Vi,Ydat[Yi], Value);
                    Exit
                  end
               end;
              if Yi < Ya - 1 then
               begin
                 Yi := Ya - 1; Vi := ValueInRowPos(Data, Xdat, X, Yi, Xtab)
               end;
              if (Vi = Va) or ((Vi > Va) = Grouth) then
               repeat
                 Vi := Va; Yi := Ya;
                 Inc(Ya); Va := ValueInRowPos(Data, Xdat, X, Ya, Xtab)
               until (Vi <> Va) and ((Vi < Va) = Grouth)
              else if (Yi > 0) and (Vi <> Value) and ((Vi > Value) = Grouth)
                 then
                  begin
                    Y := Yi - 1; V := ValueInRowPos(Data, Xdat, X, Y, Xtab);
                    while (V <> Vi) and ((V < Vi) = Grouth) do
                     begin
                       Va := Vi; Vi := V; Ya := Yi; Yi := Y;
                       Dec(Y); V := ValueInRowPos(Data, Xdat, X, Y, Xtab);
                       if (V = Value) or ((Value > V) = Grouth) then
                        begin Va := Vi; Vi := V; Ya := Yi; Yi := Y; Break end
                     end
                  end;
              if Vi = Value then Ya := Yi else if Va = Value then Yi := Ya
              else if ((Vi > Value) = Grouth) then Y := Yi-1 else Y := Yi
            end
         end
      end
   end;

  while Ya > Yi + 1 do
   begin
    Y := Round(((Va*3.0 - Vi - Value*2.0)*Yi+
                (Va - Vi*3.0 + Value*2.0)*Ya)*0.25/(Va - Vi));
    V := ValueInRowPos(Data, Xdat, X, Y, Xtab);
    if V = Value then begin Yi := Y; Ya := Y end
    else if (V < Value) = Grouth then
     begin
       Yi := Y+1; Vi := ValueInRowPos(Data, Xdat, X, Yi, Xtab);
       if Vi = Value then Ya := Yi
       else if (Vi > Value) = Grouth then
            begin Ya := Yi; Yi := Y; Va := Vi; Vi := V end
     end
    else
     begin
       Ya := Y-1; Va := ValueInRowPos(Data, Xdat, X, Ya, Xtab);
       if Va = Value then Yi := Ya
       else if (Va < Value) = Grouth then
            begin Yi := Ya; Ya := Y; Y := Yi; Vi := Va; Va := V end
     end
   end;

   if Yi = Ya then begin Ytab := Yi; Result := Ydat[Yi] end
   else
    begin
      Ytab := Y;
      if Y < Yi then Result := LinExtrapol(Vi,Ydat[Yi], Va,Ydat[Ya], Value)
      else if Y >= Ya then Result := LinExtrapol(Va,Ydat[Ya],Vi,Ydat[Yi], Value)
      else if Yi = 0 then
           Result := InvGrP3Interp(Ydat[Yi],Vi, Ydat[Ya],Va,
                          Ydat[3],ValueInRowPos(Data, Xdat, X, 3, Xtab), Value)
      else if Ya = H then
           Result := InvGrP3Interp(Ydat[Ya],Va, Ydat[Yi],Vi,
                          Ydat[H-2],ValueInRowPos(Data,Xdat,X,H-2,Xtab), Value)
      else Result := InvP4Interp(Ydat[Yi-1],ValueInRowPos(Data,Xdat,X,Yi-1,Xtab),
                                 Ydat[Yi],Vi, Ydat[Ya], Va,
                                 Ydat[Ya+1],ValueInRowPos(Data,Xdat,X,Ya+1,Xtab),
                                 Value)
    end
end;


function Xder(const Data: TReal2Array; X, Y: Extended): Extended;// dGetValue/dX
var P: Integer;
begin
  P := Floor(X);
  if P < 0 then Result := ValueInCol(Data, 1, Y) - ValueInCol(Data, 0, Y)
  else if P = 0
  then Result := dGrEq3Interp(ValueInCol(Data, 0, Y), ValueInCol(Data, 1, Y),
                              ValueInCol(Data, 2, Y), X)
  else if P < High(Data[0])-1
  then Result := dEq4Interp(ValueInCol(Data, P-1, Y), ValueInCol(Data, P, Y),
                            ValueInCol(Data, P+1, Y), ValueInCol(Data, P+2, Y),
                            X-P)
  else if P = High(Data[0])-1
  then Result := -dGrEq3Interp(ValueInCol(Data, P+1, Y), ValueInCol(Data, P, Y),
                               ValueInCol(Data, P-1, Y), P+1-X)
  else
  begin
   P := High(Data[0]); Result := ValueInCol(Data, P,Y) - ValueInCol(Data, P-1,Y)
  end
end;

function Xder(const Data: TReal2Array; const Xdat, Ydat: array of Extended;
         X, Y: Extended; var Xid, Yid: Integer): Extended; overload;
begin
  GetPos(Xdat, X, Xid); GetPos(Ydat, Y, Yid);
  if Xid < 0
  then Result := (ValueInCol(Data,Ydat,1,Y,Yid) - ValueInCol(Data,Ydat,0,Y,Yid))
                 /(Xdat[1] - Xdat[0])
  else if Xid = 0
  then Result := dGrP3Interp(Xdat[0], ValueInCol(Data, Ydat, 0, Y, Yid),
                             Xdat[1], ValueInCol(Data, Ydat, 1, Y, Yid),
                             Xdat[2], ValueInCol(Data, Ydat, 2, Y, Yid), X)
  else if Xid < High(Data[0])-1
  then Result := dP4Interp(Xdat[Xid-1],ValueInCol(Data, Ydat, Xid-1, Y, Yid),
                           Xdat[Xid], ValueInCol(Data, Ydat, Xid, Y, Yid),
                           Xdat[Xid+1],ValueInCol(Data, Ydat, Xid+1, Y, Yid),
                           Xdat[Xid+2],ValueInCol(Data, Ydat, Xid+2, Y, Yid), X)
  else if Xid = High(Data[0])-1
  then Result := -dGrP3Interp(Xdat[Xid+1],ValueInCol(Data,Ydat,Xid+1,Y,Yid),
                              Xdat[Xid], ValueInCol(Data, Ydat, Xid, Y, Yid),
                              Xdat[Xid-1],ValueInCol(Data,Ydat,Xid-1,Y,Yid), X)
  else
   Result := (ValueInCol(Data,Ydat,Xid,Y,Yid)-ValueInCol(Data,Ydat,Xid-1,Y,Yid))
             /( Xdat[Xid] - Xdat[Xid-1])
end;

function Yder(const Data: TReal2Array; X, Y: Extended): Extended;// dGetValue/dY
var P: Integer;
begin
  P := Floor(Y);
  if P < 0 then Result := ValueInRow(Data, X, 1) - ValueInRow(Data, X, 0)
  else if P = 0
  then Result := dGrEq3Interp(ValueInRow(Data, X, 0), ValueInRow(Data, X, 1),
                              ValueInRow(Data, X, 2), Y)
  else if P < High(Data)-1
  then Result := dEq4Interp(ValueInRow(Data, X, P-1), ValueInRow(Data, X, P),
                            ValueInRow(Data, X, P+1), ValueInRow(Data, X, P+2),
                            Y-P)
  else if P = High(Data)-1
  then Result := -dGrEq3Interp(ValueInRow(Data, X, P+1), ValueInRow(Data, X, P),
                               ValueInRow(Data, X, P-1), P+1-Y)
  else
  begin
   P := High(Data); Result := ValueInRow(Data, X, P) - ValueInRow(Data, X, P-1)
  end
end;

function Yder(const Data: TReal2Array; const Xdat, Ydat: array of Extended;
         X, Y: Extended; var Xid, Yid: Integer): Extended; overload;
begin
  GetPos(Xdat, X, Xid); GetPos(Ydat, Y, Yid);
  if Yid < 0
  then Result := (ValueInRow(Data,Xdat,X,1,Xid) - ValueInRow(Data,Xdat,X,0,Xid))
                 /(Ydat[1] - Ydat[0])
  else if Yid = 0
  then Result := dGrP3Interp(Ydat[0], ValueInRow(Data, Xdat, X, 0, Xid),
                             Ydat[1], ValueInRow(Data, Xdat, X, 1, Xid),
                             Ydat[2], ValueInRow(Data, Xdat, X, 2, Xid), Y)
  else if Yid < High(Data)-1
  then Result := dP4Interp(Ydat[Yid-1],ValueInRow(Data, Xdat, X, Yid-1, Xid),
                           Ydat[Yid], ValueInRow(Data, Xdat, X, Yid, Xid),
                           Ydat[Yid+1],ValueInRow(Data, Xdat, X, Yid+1, Xid),
                           Ydat[Yid+2],ValueInRow(Data, Xdat, X, Yid+2, Xid), Y)
  else if Xid = High(Data[0])-1
  then Result := -dGrP3Interp(Ydat[Yid+1],ValueInRow(Data,Xdat,X,Yid+1,Xid),
                              Ydat[Yid], ValueInRow(Data, Xdat, X, Yid, Xid),
                              Ydat[Yid-1],ValueInRow(Data,Xdat,X,Yid-1,Xid), Y)
  else
   Result := (ValueInRow(Data,Xdat,X,Yid,Xid)-ValueInRow(Data,Xdat,X-1,Yid,Xid))
             /(Ydat[Yid] - Ydat[Yid-1])
end;

end.
